from __future__ import annotations

import os
from collections.abc import Callable
from concurrent.futures import ProcessPoolExecutor
from functools import partial

import numpy as np
from ribs.optimizers import Optimizer
from sklearn.utils import check_random_state

from phem.base_utils.metrics import AbstractMetric
from phem.framework.abstract_weighted_ensemble import AbstractWeightedEnsemble
from phem.methods.ensemble_selection.qdo.behavior_space import BehaviorSpace
from phem.methods.ensemble_selection.qdo.custom_archives.custom_sliding_boundaries_archive import (
    SlidingBoundariesArchive,
)
from phem.methods.ensemble_selection.qdo.custom_archives.quality_archive import QualityArchive
from phem.methods.ensemble_selection.qdo.emitters import DiscreteWeightSpaceEmitter


class QDOEnsembleSelection(AbstractWeightedEnsemble):
    """Using QDO to find a good weight vector for post hoc ensembling.

    Parameters
    ----------
    base_models: List[Callable], List[sklearn estimators]
        The pool of fitted base models.
    n_iterations: int
        The number of iterations determines the number of evaluations. By default, number of evaluations
        = n_iterations * n_base_models. The true number of iterations is also determined by the batch size.
        If n_base_models == batch_size, then used n_iterations is the value that is passed. Otherwise, we need to adapt
        the number of iterations.
            FIXME: change stopping criterion based on time or number of evaluations instead of iterations.
    score_metric: AbstractMetric
        The metric used to evaluate the models' performance.
    behavior_space: BehaviorSpace, default=None
        The behavior space used to determine the behavior of a solution. If None, we assume that
        the framework is used for quality search and we supply a dummy behavior space.
    archive_type: str in {"sliding", "quality"}, default="sliding"
        Defines which archive type to use (i.e., how the bins are set up).
        *   "sliding": Sliding Boundaries Archive; remaps bounds of each dimension in set intervals.
            We use this as default as diversity values in ensembling have no guarantee to be uniformly distributed
            in the boundaries of the values. This is usually the reason why one uses Grid or CVT based archives.
        *   "quality": Traditional population based search without diversity. To do so, we use an archive that
            represents our population of good solutions. That is, a collection of top N elites based on objective value.
            Consequently, any behavior values computed by the behavior space are ignored. "quality" can be used as a
            baseline to see if diversity helps at all.
    buffer_ratio: float, default=1.
        Only relevant for sliding-based archives. The size of the buffer relative to the number of seen solutions.
        For example, 1. means we use all seen solutions for remaps; with 0.5, we use a buffer of size 50% of the total
        solutions. It must hold that 0 < buffer_ratio <= 1.
    max_elites: int, default=49
        Defines the archive size (number of bins). Must be a number that we can split evenly by the number of
        behavior dimensions if we use archives (e.g. a square, a cube, ...).
    batch_size: int, default=40
        Defines the batch size for ask/tell loop.
    emitter_method: str in {"DiscreteWeightSpaceEmitter"}, default="DiscreteWeightSpaceEmitter"
        Defined which hard-coded configuration you want to use for the emitter.
        *   DiscreteWeightSpaceEmitter: An emitter working in the discrete weight space.
    emitter_initialization_method: str in
                {"AllL1", "RandomL2Combinations", "L2ofSingleBest"}, default="RandomL2Combinations"
        Defines the initialization method for the first batch of solutions. That is, the set of initial weight
        vectors.
        *   "AllL1": The first batch of solutions are all base models. That is, the weight vector is a one-hot vector.
        *   "RandomL2Combinations": The first batch of solutions contains random combination of base models.
            As a result, it will start with ensembles of size 2, i.e., Layer 2 (L2). The combinations are done
            exhaustively such that we will produce n_base_models/2 many solutions.
        *   "L2ofSingleBest": Extend the single best model to an L2 ensemble for all possible combinations. Produces
            n_base_models-1 many initial solutions.
    emitter_vars: dict, default=None
        Optional variables passed to the emitter. See emitter documentation to find out what you can pass here.
    random_state: Optional[int | RandomState] = None
        The random_state used for ensemble selection.
        *   None - Uses numpy's default RandomState object
        *   int - Successive calls to fit will produce the same results
        *   RandomState - Truely random, each call to fit will produce
                          different results, even with the same object.
    n_jobs: int, default=-1
        Cores to use for parallelization. If -1, use all available cores.
        Please be aware that multi-processing introduces a time overhead.
    show_analysis: bool, default=False
        Show fit analysis code.
    """

    allowed_emitter_methods = {"DiscreteWeightSpaceEmitter"}
    allowed_emitter_initialization_methods = {"AllL1", "RandomL2Combinations", "L2ofSingleBest"}

    def __init__(
        self,
        base_models: list[Callable],
        n_iterations: int,
        score_metric: AbstractMetric,
        behavior_space: BehaviorSpace | None = None,
        archive_type: str = "sliding",
        buffer_ratio: float = 1.0,
        max_elites: int = 49,
        emitter_vars: dict | None = None,
        batch_size: int = 40,
        emitter_method: str = "DiscreteWeightSpaceEmitter",
        emitter_initialization_method: str = "RandomL2Combinations",
        random_state: int | np.random.RandomState | None = None,
        show_analysis=False,
        n_jobs: int = -1,
    ) -> None:
        # The following just tells our wrapper/supper that we want to have prediction probabilities as input for fit
        super().__init__(base_models, "predict_proba")

        # -- Iteration management
        self.n_iterations = n_iterations
        self.n_base_models = len(base_models)
        self.batch_size = batch_size
        self.n_evaluations = self.n_iterations * self.n_base_models
        self.internal_n_iterations, self.n_rest_evaluations = self._compute_n_iterations(
            self.n_evaluations, self.batch_size
        )

        # -- Basic Init
        self.archive_type = archive_type
        self.max_elites = max_elites
        self.emitter_method = emitter_method
        self.show_analysis = show_analysis
        self.buffer_ratio = buffer_ratio
        self.random_state = check_random_state(random_state)

        # -- Metrics / Behavior
        self.score_metric = score_metric
        self.behavior_space = behavior_space if behavior_space is not None else BehaviorSpace([])

        # -- More Advanced
        self.emitter_vars = {} if emitter_vars is None else emitter_vars
        self.init_iteration_results = None  # type: Optional[List[float, np.ndarray]]
        self._res_buffer = np.ndarray((0, 1 + len(self.behavior_space.behavior_functions)))
        self.emitter_initialization_method = emitter_initialization_method
        self._n_init_evals = 0

        # -- Pre-process metadata from base models
        self.config_key_to_range_ = None  # type: Optional[Dict[str, float]]
        self.qdo_base_models_metadata_ = None  # type: Optional[List[dict]]
        if self.behavior_space.requires_base_model_metadata:
            self.qdo_base_models_metadata_ = []  # Only consists of the config and no other metadata
            for bm_metadata in self.base_models_metadata:
                if not (bm_metadata.get("auto-sklearn-model")):
                    raise NotImplementedError(
                        "We currently only support base model metadata created by the ",
                        "Auto-sklearn Assembler.",
                    )
                self.qdo_base_models_metadata_.append(bm_metadata["config"])

            # -- Compute ranges
            all_numeric_keys = [
                (k, v)
                for bm in self.base_models_metadata
                for k, v in bm["config"].items()
                if not isinstance(v, str)
            ]
            # Get bounds
            key_to_bounds = {
                k: [float("+inf"), float("-inf")] for k in {k for k, v in all_numeric_keys}
            }
            for k, v in all_numeric_keys:
                curr_min = key_to_bounds[k][0]
                curr_max = key_to_bounds[k][1]
                if v < curr_min:
                    key_to_bounds[k][0] = v
                if v > curr_max:
                    key_to_bounds[k][1] = v

            # Transform to ranges
            self.config_key_to_range_ = {
                k: abs(bds[1] - bds[0]) for k, bds in key_to_bounds.items()
            }

        # -- Multi-processing
        if (n_jobs == 1) or (os.name == "nt"):
            self._use_mp = False
            if os.name == "nt":
                pass
        else:
            if n_jobs == -1:
                n_jobs = len(os.sched_getaffinity(0))
            self._n_jobs = n_jobs
            self._use_mp = True

    def ensemble_fit(
        self, predictions: list[np.ndarray], labels: np.ndarray
    ) -> AbstractWeightedEnsemble:
        # -- Input Validation
        self.n_iterations = int(self.n_iterations)
        if self.n_iterations < 1:
            raise ValueError("Number of evaluations cannot be less than one!")

        if not isinstance(self.score_metric, AbstractMetric):
            raise ValueError(
                "The provided metric must be an instance of a AbstractMetric, "
                f"nevertheless it is {self.score_metric}({type(self.score_metric)})"
            )

        if not isinstance(self.behavior_space, BehaviorSpace):
            raise ValueError(
                "The provided behavior space must be an instance of a BehaviorSpace, "
                f"nevertheless it is {self.behavior_space}({type(self.behavior_space)})"
            )

        if self.emitter_method not in self.allowed_emitter_methods:
            raise ValueError(
                f"The emitter methods is not in {self.allowed_emitter_methods}. Got: {self.emitter_method}"
            )

        if self.emitter_initialization_method not in self.allowed_emitter_initialization_methods:
            raise ValueError(
                f"The emitter initialization method is not in {self.allowed_emitter_initialization_methods}. Got: {self.emitter_initialization_method}"
            )

        if (not isinstance(self.buffer_ratio, float)) or (
            isinstance(self.buffer_ratio, float)
            and ((self.buffer_ratio <= 0) or (self.buffer_ratio > 1))
        ):
            raise ValueError(
                "Buffer ratio must be a float larger zero and smaller or equal to 1.",
                f"Got: {self.buffer_ratio}, {type(self.buffer_ratio)}",
            )

        if (self.archive_type not in ["quality"]) and int(
            round(self.max_elites ** (1.0 / self.behavior_space.n_dims))
        ) ** self.behavior_space.n_dims != self.max_elites:
            raise ValueError(
                "Number of max elites can not be used to split the dimensions evenly!",
                f"Got: {self.max_elites}, {self.behavior_space.n_dims}.",
            )

        # -- Init Prerequisites
        self._init_optimize(predictions)

        # -- Init QDO Elements
        self._init_archive()
        self._init_emitters(predictions, labels)

        # -- Call optimizer
        self._optimize(predictions, labels)

        # -- Get final weights
        self._compute_weights(predictions, labels)

        if self.show_analysis:
            self._analysis_fit(predictions, labels)

        return self

    def _init_archive(self):
        # -- Build Archive
        if self.archive_type == "sliding":
            archive = SlidingBoundariesArchive(
                dims=[
                    int(round(self.max_elites ** (1.0 / self.behavior_space.n_dims)))
                    for _ in range(self.behavior_space.n_dims)
                ],
                ranges=self.behavior_space.ranges,
                seed=self.random_state.randint(1, 100000),
                remap_frequency=self.n_base_models,
                initial_remap=self.n_base_models + self.batch_size,
                buffer_capacity=int((self.n_evaluations + 10) * self.buffer_ratio),
                show_analysis=self.show_analysis,
            )
        elif self.archive_type == "quality":
            archive = QualityArchive(
                archive_size=self.max_elites,
                behavior_n_dim=self.behavior_space.n_dims,
                seed=self.random_state.randint(1, 100000),
                show_analysis=self.show_analysis,
            )
        else:
            raise ValueError("Unknown Archive Type!")

        self.archive = archive

    def _update_n_iterations(self, n_evals_outside_of_loop: int):
        # Set parameter for logging to see how many values were evaluated outside the loop
        self._n_init_evals += n_evals_outside_of_loop
        # Set number of iterations to reflect this and to guarantee that we do not exceed the allowed
        # number of evaluations
        self.internal_n_iterations, self.n_rest_evaluations = self._compute_n_iterations(
            self.n_evaluations - self._n_init_evals, self.batch_size
        )

    @staticmethod
    def _compute_n_iterations(n_eval, batch_size):
        internal_n_iterations = n_eval // batch_size

        n_rest_evaluations = 0 if n_eval % batch_size == 0 else n_eval % batch_size

        return internal_n_iterations, n_rest_evaluations

    def _init_start_weight_vectors(self, predictions: list[np.ndarray], y_true: np.ndarray):
        """Initialize the weight vectors for the first batch of solutions to be evaluated.

        We must set self.init_iteration_results if the start weight vectors do not include the Single Best.
        """
        # -- Evaluate Base Models and to determine start weight vectors
        base_weight_vector = np.zeros((self.n_base_models,))
        _start_weight_vectors = []

        # -- Get L1 Weight Vectors
        for i in range(self.n_base_models):
            tmp_w = base_weight_vector.copy()
            tmp_w[i] = 1
            _start_weight_vectors.append((tmp_w, 1))

        # -- Switch case for different initialization methods
        if self.emitter_initialization_method != "AllL1":
            # -- Get Single Best to guarantee performance improvement based on validation data
            objs, _ = self._evaluate_batch_of_solutions(
                np.array([w for w, _ in _start_weight_vectors]), predictions, y_true
            )
            sb_weight_vector = _start_weight_vectors[np.argmax(objs)]
            self._update_n_iterations(self.n_base_models)

            # -- Switch case
            if self.emitter_initialization_method == "RandomL2Combinations":
                self.init_iteration_results = (np.max(objs), sb_weight_vector[0])

                l2_combinations = self.random_state.choice(
                    self.n_base_models, (self.n_base_models // 2, 2), replace=False
                )
                # Fill combinations
                _start_weight_vectors = []
                for i, j in l2_combinations:
                    tmp_w = base_weight_vector.copy()
                    tmp_w[i] = 0.5
                    tmp_w[j] = 0.5
                    _start_weight_vectors.append((tmp_w, 2))

            elif self.emitter_initialization_method == "L2ofSingleBest":
                self.init_iteration_results = (np.max(objs), sb_weight_vector[0])
                index_single_best = np.argmax(objs)

                # Fill combinations
                _start_weight_vectors = []
                for i in range(self.n_base_models):
                    if i == index_single_best:
                        continue
                    tmp_w = base_weight_vector.copy()
                    tmp_w[i] = 0.5
                    tmp_w[index_single_best] = 0.5
                    _start_weight_vectors.append((tmp_w, 2))

        # If it is "AllL1" we do need to anything and just return all of them
        return _start_weight_vectors

    def _init_emitters(self, predictions, y_true):
        """Init emitter.

        Step Emitters does most of the initialization internally.

        The following must be set in the init:
        ---------
        self.discrete_merge_ : bool
            True if we need to make weights discrete for merging at the end during
            compute weights. Otherwise, false.
        self.init_iteration_results: Optional[List[float, np.ndarray]]
            Set to the best initial result; needed if the archive does not evaluate
            the initial solutions (e.g. x0). Otherwise, validation score could be
            worse than the single best.
        """
        _start_weight_vectors = self._init_start_weight_vectors(predictions, y_true)

        # Code for analysis
        if self.show_analysis:
            o, b = self._evaluate_batch_of_solutions(
                np.array([w for w, _ in _start_weight_vectors]), predictions, y_true
            )
            self._sb_stats = [o[np.argmax(o)], b[np.argmax(o), :]]

        # -- Build emitters
        if self.emitter_method == "DiscreteWeightSpaceEmitter":
            for req_time_atr in ["mutation_probability_after_crossover", "crossover_probability"]:
                if (req_time_atr in self.emitter_vars) and (
                    isinstance(self.emitter_vars[req_time_atr], Callable)
                ):
                    self.emitter_vars[req_time_atr] = partial(
                        self.emitter_vars[req_time_atr], max_time=self.internal_n_iterations
                    )

            emitters = [
                DiscreteWeightSpaceEmitter(
                    self.archive,
                    self.n_base_models,
                    _start_weight_vectors,
                    batch_size=self.batch_size,
                    seed=self.random_state.randint(1, 100000),
                    **self.emitter_vars,
                ),
            ]
            self.discrete_merge_ = True
        else:
            raise RuntimeError("You should not be here!")

        self.emitters = emitters

        # Set proba normalization if needed
        #   If negative weights are proposed, we need to normalize probabilities
        #   for some metrics. Here, we set a flag to do so later.
        metrics_need_proba = self.score_metric.requires_confidences or (
            "proba" in self.behavior_space.required_prediction_types
        )
        if any(emt.proposes_negative_weights for emt in self.emitters) and metrics_need_proba:
            self.normalize_predict_proba_ = True

    def _init_optimize(self, predictions):
        # Pre-compute values that are regularly needed for evaluation

        self._raw_predictions = None  # type: Optional[np.ndarray]

        if "raw" in self.behavior_space.required_prediction_types:
            self._raw_predictions = [self._confidences_to_predictions(bm) for bm in predictions]

    def _optimize(self, predictions: list[np.ndarray], labels: np.ndarray) -> None:
        # -- Build Optimizer
        opt = Optimizer(self.archive, self.emitters)

        # -- Set up Results collector
        optimize_stats = {
            "Archive Size": [],
            "Max Objective": [],
        }
        if self.init_iteration_results is not None:
            optimize_stats["Archive Size"].append(0)
            optimize_stats["Max Objective"].append(self.init_iteration_results[0])

        for _itr in range(1, self.internal_n_iterations + 1):
            # Get solutions
            sols = opt.ask()

            # Evaluate solutions
            objs, bcs = self._evaluate_batch_of_solutions(sols, predictions, labels)

            # Report back and restart
            opt.tell(objs, bcs)

            # Log stats
            optimize_stats["Archive Size"].append(len(opt.archive))
            optimize_stats["Max Objective"].append(float(opt.archive.stats.obj_max))

        # -- Rest Iteration, required if wanted number of evaluations can not be evenly
        #    distributed across batches.
        if self.n_rest_evaluations:
            sols = opt.ask()
            org_length = len(sols)
            sols = sols[: self.n_rest_evaluations, :]

            objs, bcs = self._evaluate_batch_of_solutions(sols, predictions, labels)

            # Get some existing behavior values and a worse objective value
            dummy_obj = objs[0]
            dummy_bc = bcs[0, :]

            # fill rest of the solutions with dummy values
            objs = np.hstack([objs, [dummy_obj] * (org_length - len(sols))])
            bcs = np.vstack([bcs, [dummy_bc] * (org_length - len(sols))])

            opt.tell(objs, bcs)
            optimize_stats["Archive Size"].append(len(opt.archive))
            optimize_stats["Max Objective"].append(float(opt.archive.stats.obj_max))

        self.optimize_stats_ = optimize_stats

    def _compute_weights(self, predictions, labels):
        """Code to compute the final weight vector (among other things).

        Code does the following:
            1. First it tests if merging all found solutions is better than the single
                best solution. If yes, the merged solution is used.
            2. Second if verifies (if needed) if the best found solution improved
                over the single best base model and returns the single best
                base model's results.
            3. Lastly, it fills metadat about the training process such that we can analysis them later.
        """
        # -- Get Best Weight Vector
        elites = list(self.archive)
        performances = [e.obj for e in elites]

        # - Get merged weights
        if self.discrete_merge_:
            disc_weights = np.array([elite.sol * elite.meta for elite in elites])
            merged_weights = np.sum(disc_weights, axis=0) / np.sum(np.sum(disc_weights, axis=1))
        else:
            merged_weights = np.array([elite.sol for elite in elites]).sum(axis=0) / len(elites)
        # Get performance for merged weights
        merge_obj = self._evaluate_batch_of_solutions(
            np.array([merged_weights]), predictions, labels
        )[0][0]

        # max/Argmax because we made this a maximization problem to work with ribs
        if merge_obj > np.max(performances):
            self.optimize_stats_["merged_weights"] = True
            self.weights_ = merged_weights
            self.validation_loss_ = -float(merge_obj)

        else:
            self.optimize_stats_["merged_weights"] = False
            self.weights_ = elites[np.argmax(performances)].sol
            self.validation_loss_ = -float(np.max(performances))

        # -- Verify that the optimization method improved over the single best
        #   Only done if the method does/can not do this by itself
        if self.init_iteration_results is not None:
            init_score, init_weights = self.init_iteration_results

            # >= because ini iteration has most likely a smaller ensemble size
            if self.validation_loss_ >= -float(init_score):
                self.optimize_stats_["merged_weights"] = False
                self.weights_ = init_weights
                self.validation_loss_ = -float(init_score)

        # -- Set to save metadata
        self.iteration_batch_size_ = self.batch_size
        self.val_loss_over_iterations_ = [-i for i in self.optimize_stats_["Max Objective"]]

        # - set additional metadata
        add_var = isinstance(self.emitters[0], DiscreteWeightSpaceEmitter)
        self.model_specific_metadata_ = {
            "evaluation_types": {
                "total": int(self.n_base_models * self.n_iterations),
                "explore": sum(int(em.explore) for em in self.emitters) if add_var else -1,
                "exploit": sum(int(em.exploit) for em in self.emitters) if add_var else -1,
                "init": int(self._n_init_evals) if add_var else -1,
                "rejects": sum(int(em._total_rejects) for em in self.emitters) if add_var else -1,
                "crossover_rejects": sum(int(em._total_crossover_rejects) for em in self.emitters)
                if add_var
                else -1,
                "n_mutate": sum(int(em.n_mutate) for em in self.emitters) if add_var else -1,
                "n_crossover": sum(int(em.n_crossover) for em in self.emitters) if add_var else -1,
            },
            "internal_n_iterations": int(self.internal_n_iterations)
            + int(self.n_rest_evaluations > 0),
            "archive_size": [int(i) for i in self.optimize_stats_["Archive Size"]],
        }

    def _evaluate_batch_of_solutions(self, solutions: np.ndarray, predictions, y_true):
        """Return objective value and BC for a batch of solutions.

        Parameters
        ----------
        solutions: np.ndarray (batch_size, n_base_models)
            A batch of weight vectors.

        Returns:
        -------
            objs (np.ndarray): (batch_size,) array with objective values.
            bcs (np.ndarray): (batch_size,) array with a BC in each row.
        """
        # Determine function arguments
        input_required = self.behavior_space.required_prediction_types
        raw_req = "raw" in input_required
        proba_req = "proba" in input_required
        bm_meta_req = self.behavior_space.requires_base_model_metadata

        # Create static function arguments list
        func_args = [
            predictions,
            y_true,
            self.score_metric,
            self.behavior_space,
            self.normalize_predict_proba_,
            proba_req,
            raw_req,
            self._raw_predictions,
            bm_meta_req,
            self.config_key_to_range_,
            self.qdo_base_models_metadata_,
        ]

        if self._use_mp:
            func_args.append(solutions)
            sol_i_list = list(range(len(solutions)))

            with ProcessPoolExecutor(
                self._n_jobs, initializer=_pool_init, initargs=func_args
            ) as ex:
                results = ex.map(_init_wrapper_evaluate_single_solution, sol_i_list)
            res = np.array(list(results))
        else:
            res = np.apply_along_axis(
                partial(evaluate_single_solution, *func_args), axis=1, arr=solutions
            )

        if self.show_analysis:
            self._res_buffer = np.vstack([self._res_buffer, res])

        return res[:, 0], res[:, 1:]

    def _analysis_fit(self, predictions, labels, plot=False, plot_animation=False):
        """Some basic stuff to look at the result of fit."""
        raise NotImplementedError()


def _pool_init(
    _predictions,
    _y_true,
    _score_metric,
    _behavior_space,
    _normalize_predict_proba_,
    _proba_req,
    _raw_req,
    _raw_predictions,
    _bm_meta_req,
    _config_key_to_range_,
    _qdo_base_models_metadata_,
    _solutions,
):
    global \
        predictions, \
        y_true, \
        score_metric, \
        behavior_space, \
        normalize_predict_proba_, \
        proba_req, \
        raw_req, \
        raw_predictions, \
        bm_meta_req, \
        config_key_to_range_, \
        qdo_base_models_metadata_, \
        solutions

    predictions = _predictions
    y_true = _y_true
    score_metric = _score_metric
    behavior_space = _behavior_space
    normalize_predict_proba_ = _normalize_predict_proba_
    proba_req = _proba_req
    raw_req = _raw_req
    raw_predictions = _raw_predictions
    bm_meta_req = _bm_meta_req
    config_key_to_range_ = _config_key_to_range_
    qdo_base_models_metadata_ = _qdo_base_models_metadata_
    solutions = _solutions


def _init_wrapper_evaluate_single_solution(sol_index):
    return evaluate_single_solution(
        predictions,
        y_true,
        score_metric,
        behavior_space,
        normalize_predict_proba_,
        proba_req,
        raw_req,
        raw_predictions,
        bm_meta_req,
        config_key_to_range_,
        qdo_base_models_metadata_,
        solutions[sol_index],
    )


def evaluate_single_solution(
    predictions,
    y_true,
    score_metric,
    behavior_space,
    normalize_predict_proba_,
    proba_req,
    raw_req,
    raw_predictions,
    bm_meta_req,
    config_key_to_range_,
    qdo_base_models_metadata_,
    weight_vector,
):
    """Static Out-of-Class Function to avoid pickle problems in multiprocessing.

    Parameters
    ----------
    predictions: (see ensemble_fit)
    y_true: (see ensemble_fit)
    score_metric: AbstractMetric
        Metric used to score the solution
    behavior_space: BehaviorSpace
        Behavior Space used to evaluate the diversity of the solution
    normalize_predict_proba_: bool
        Required for ensemble_predict to know if the proba scores need to be normalized
    proba_req: bool
        Whether the behavior space needs proba predictions
    raw_req: bool
        Whether the behavior space needs raw predictions
    raw_predictions: (see ensemble_fit and init_optimize)
        The pre-computed raw predictions of all base models. Only required if raw_req is True.
    bm_meta_req: bool
        Whether the behavior space needs base mode metadata
    config_key_to_range_: (see class init)
        Base Mode metadata. Only required if bm_meta_req is True.
    qdo_base_models_metadata_: (see class init)
        Base Mode metadata. Only required if bm_meta_req is True.
    weight_vector: array-like, shape (n_base_models,)
        A specific solution vector that is to be evaluated.

    Returns:
    -------
        s_m: Float
            Score of Solution (as loss)
        bcs: array-like, shape (behavior_space.n_dims,)
            The behavior function values of the solution.
    """
    # Get Score
    y_pred_ensemble = AbstractWeightedEnsemble._ensemble_predict(
        predictions, weight_vector, normalize_predict_proba_
    )

    # Negative loss because we want to maximize
    s_m = -score_metric(y_true, y_pred_ensemble, to_loss=True, checks=False)

    # Base Model Y_preds
    rm_zero_mask = weight_vector != 0
    Y_pred_base_models = [
        bm for bm, selected in zip(predictions, rm_zero_mask, strict=False) if selected
    ]

    # Handle arguments
    pred_arguments = {}
    if raw_req:
        pred_arguments["raw_preds"] = [
            AbstractWeightedEnsemble._confidences_to_predictions(y_pred_ensemble),
            [bm for bm, selected in zip(raw_predictions, rm_zero_mask, strict=False) if selected],
        ]

    if proba_req:
        pred_arguments["proba_preds"] = [y_pred_ensemble, Y_pred_base_models]

    if bm_meta_req:
        pred_arguments["input_metadata"] = [
            config_key_to_range_,
            [
                md
                for md, selected in zip(qdo_base_models_metadata_, rm_zero_mask, strict=False)
                if selected
            ],
        ]

    b_i = behavior_space(weight_vector, y_true, **pred_arguments) if pred_arguments else []

    return s_m, *b_i



"""Code for behavior space.

Example:
----------
    # --- Space for Behavior space with two diversity metrics
    bs = BehaviorSpace([
        BehaviorFunction(partial(ErrorCorrelation, checks=False),
                         ["y_true", "Y_pred_base_models"],
                         (0, 1),
                         "proba"),
        BehaviorFunction(partial(Q_statistic, checks=False),
                         ["y_true", "Y_pred_base_models"],
                         (0, 1),
                         "raw")
    ])

"""
from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as np

ALLOWED_ARGUMENTS = ["y_true", "y_pred_ensemble", "Y_pred_base_models", "weights",
                     "input_metadata", "y_pred"]


# --- Class
class BehaviorFunction:
    """Behavior function class.

    Parameters
    ----------
    function: Callable function returning a float
        Takes as input the arguments specified in  required_arguments and returns a float representing
        the behavior
    required_arguments: List of str
        The required arguments are supposed to be a list of strings defining which arguments are required and in
        which order. Assume that the available arguments' name are identical to the parameter names in the function.
            FIXME: add option to allow alternative parameter names in function
        Available arguments are: {"y_true", "y_pred_ensemble", "Y_pred_base_models", "weights", "base_models_metadata"}
    range_tuple: Tuple of floats
        Define the upper and lower bound of this behavior function
    required_prediction_format: str in {"raw", "proba", "none"}
        Defines in which format the predictions are passed to the function, if at all.
    name: str, default=None
        A name of the behavior function.
    """

    def __init__(self, function: Callable[[], float], required_arguments: list[str],
                 range_tuple: tuple[float, float], required_prediction_format: str, name: str | None = None):
        self.function = function

        if any(v not in ALLOWED_ARGUMENTS for v in required_arguments):
            raise ValueError(f"Not allowed argument name used. Allowed are: {ALLOWED_ARGUMENTS}. Got: {required_arguments}")
        self.required_arguments = required_arguments
        self.range_tuple = range_tuple

        if required_prediction_format not in ["raw", "proba", "none"]:
            raise ValueError("Unknown prediction format. Expected: {}. Got: .{}".format(["raw", "proba", "none"],
                                                                                        required_prediction_format))
        self.required_prediction_format = required_prediction_format
        self.name = "Placeholder" if name is None else name
        self.requires_base_model_metadata = "input_metadata" in required_arguments


class BehaviorSpace:
    """A class for behavior functions for QDO."""

    def __init__(self, behavior_functions: list[BehaviorFunction]):
        self.behavior_functions = behavior_functions

    @property
    def ranges(self):
        return [bf.range_tuple for bf in self.behavior_functions]

    @property
    def n_dims(self):
        return len(self.behavior_functions)

    @property
    def required_prediction_types(self):
        return {bf.required_prediction_format for bf in self.behavior_functions}

    @property
    def requires_base_model_metadata(self):
        return any(bf.requires_base_model_metadata for bf in self.behavior_functions)

    def __call__(self, weights: np.ndarray, y_true: np.ndarray,
                 raw_preds: tuple[np.ndarray, list[np.ndarray]] | None = (None, None),
                 proba_preds: tuple[np.ndarray, list[np.ndarray]] | None = (None, None),
                 input_metadata: Any | None = None) -> list[float]:
        """Get an instance of the behavior space.

        Parameters
        ----------
        y_true: ndarray (n_samples,)
            Ground truth / target vector
        raw_preds, proba_preds: Tuple containing the following for either raw ore proba predictions
            y_pred_ensemble: ndarray (n_samples, ) or (n_samples, n_classes)
                Contains the (probability) predictions of the ensemble
            Y_pred_base_models: List[ndarray] (n_base_models, n_samples) or (n_base_models, n_samples, n_classes)
                Contains the (probability) predictions of each base model in the ensemble
        weights: ndarray (n_base_models, ), default=None
            The weights used to compute the weighted average.
        input_metadata: List[dict], default=None
            The metadata for the metric if required.

        Returns:
        ----------
        behavior_space_instance: np.ndarray
            Float values for each dimension representing the behavior in that dimension.
        """
        raw_y_pred_ensemble, raw_Y_pred_base_models = raw_preds
        proba_y_pred_ensemble, proba_Y_pred_base_models = proba_preds

        b_space_instance = []

        for bf in self.behavior_functions:
            # Can ignore the "none" case as in that case the arguments are not used.
            requires_raw = bf.required_prediction_format == "raw"

            potential_args = {"y_true": y_true,
                                  "y_pred": raw_y_pred_ensemble if requires_raw else proba_y_pred_ensemble,
                                  "y_pred_ensemble": raw_y_pred_ensemble if requires_raw else proba_y_pred_ensemble,
                                  "Y_pred_base_models": raw_Y_pred_base_models if requires_raw else proba_Y_pred_base_models,
                                  "weights": weights,
                                  "input_metadata": input_metadata}

            b_space_instance.append(bf.function(**{para: potential_args[para] for para in bf.required_arguments}))

        return b_space_instance



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Callable

import numpy as np
from sklearn.preprocessing import LabelBinarizer
from sklearn.utils.validation import _check_y, check_array


# --- Class Stuff
class DiversityMetric:
    """An abstract class for metrics to measure diversity."""
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, metric_func, metric_name, requires_weights=False, requires_y_ensemble_pred=False,
                 requires_raw_predictions=False, more_diversity_if_higher=False, single_model_ensemble_default=1.):
        self.metric_func = metric_func
        self.name = metric_name
        self.requires_weights = requires_weights
        self.requires_y_ensemble_pred = requires_y_ensemble_pred
        self.y_ensemble_pred_signature_name = "y_pred_ensemble"
        self.weights_signature_name = "weights"
        self.requires_raw_predictions = requires_raw_predictions
        self.more_diversity_if_higher = more_diversity_if_higher
        self.single_model_ensemble_default = single_model_ensemble_default

    def __call__(self, y_true: np.ndarray, Y_pred_base_models: list[np.ndarray],
                 y_pred_ensemble: np.ndarray = None, weights: np.ndarray = None, checks=True,
                 transform_confidences: bool = False):
        """Call which determines what to pass to the metric function.

        Moreover, we define the diversity to be 1 if the ensemble consists of only one model.
        In our case, lower scores of the metric always correspond to more diverse ensembles.
        And 1 means the least amount of diversity.

        Parameters
        ----------
        Y_pred_base_models: List[ndarray] (n_base_models, n_samples) or (n_base_models, n_samples, n_classes)
            Contains the (probability) predictions of each base model in the ensemble
        y_true: ndarray (n_samples,)
            Ground truth / target vector
        y_pred_ensemble: ndarray (n_samples, ) or (n_samples, n_classes)
            Contains the (probability) predictions of the ensemble
        weights: ndarray (n_base_models, ), default=None
            The weights used to compute the weighted average.
        """
        if len(Y_pred_base_models) == 1:
            return self.single_model_ensemble_default

        if checks:
            # -- Input validation
            y_true = _check_y(y_true)

        # -- Build Kwargs that the metric needs
        kwargs = {}

        # - Handle Weight
        if self.requires_weights:
            if weights is None:
                raise ValueError("Metric requires weights but weights are None.")
            if len(weights) != len(Y_pred_base_models) and sum(weights != 0) != len(Y_pred_base_models):
                raise ValueError("Weight vector and base model count do not match!")

            kwargs[self.weights_signature_name] = weights

        if self.requires_y_ensemble_pred:
            y_pred_ensemble = check_array(y_pred_ensemble)
            kwargs[self.y_ensemble_pred_signature_name] = y_pred_ensemble

        # -- Pass data to metric
        if isinstance(self, OrdinalClassificationDiversityMetric):

            if checks:
                # Verify Dimensions
                if (y_pred_ensemble is not None) and y_pred_ensemble.ndim != 2:
                    raise ValueError("OrdinalClassificationDiversityMetric requires prediction probabilities as input! "
                                     "Ensemble Predictions (y_pred_ensemble) has more or less than 2 dimensions.")

                if any(Y_pred_base_model.ndim != 2 for Y_pred_base_model in Y_pred_base_models):
                    raise ValueError("OrdinalClassificationDiversityMetric requires prediction probabilities as input! "
                                     "The predictions of at least one base model in Y_pred_base_models is not 2 dimensional",
                                     )

                if (y_pred_ensemble is not None) \
                        and any(bm.shape[1] < y_pred_ensemble.shape[1] for bm in Y_pred_base_models):
                    raise ValueError(
                        "y_true has more classes than some base model has predicted for.")

                if (y_pred_ensemble is not None) \
                        and any(bm.shape[1] != y_pred_ensemble.shape[1] for bm in Y_pred_base_models):
                    raise ValueError(
                        "The number of classes does not match between ensemble and some base model predictions.")

                # Verify Array
                Y_pred_base_models = [check_array(Y_pred_base_model) for Y_pred_base_model in Y_pred_base_models]

            # Get Score
            diversity_score = self.metric_func(y_true, Y_pred_base_models, **kwargs)

        elif isinstance(self, NonOrdinalClassificationDiversityMetric):
            if checks:
                Y_pred_base_models = [_check_y(Y_pred_base_model) for Y_pred_base_model in Y_pred_base_models]

            diversity_score = self.metric_func(y_true, Y_pred_base_models, **kwargs)

        else:
            raise NotImplementedError()

        return diversity_score


class OrdinalClassificationDiversityMetric(DiversityMetric):

    def __init__(self, metric_func, metric_name, requires_weights, requires_y_ensemble_pred, more_diversity_if_higher,
                 single_model_ensemble_default):
        super().__init__(metric_func, metric_name,
                         requires_weights=requires_weights, requires_y_ensemble_pred=requires_y_ensemble_pred,
                         requires_raw_predictions=False, more_diversity_if_higher=more_diversity_if_higher,
                         single_model_ensemble_default=single_model_ensemble_default)


class NonOrdinalClassificationDiversityMetric(DiversityMetric):

    def __init__(self, metric_func, metric_name, requires_weights, requires_y_ensemble_pred, more_diversity_if_higher,
                 single_model_ensemble_default):
        super().__init__(metric_func, metric_name,
                         requires_weights=requires_weights, requires_y_ensemble_pred=requires_y_ensemble_pred,
                         requires_raw_predictions=True, more_diversity_if_higher=more_diversity_if_higher,
                         single_model_ensemble_default=single_model_ensemble_default)


# --- Metric Usability Stuff
def make_diversity_metric(metric_type: str, metric_name: str, metric_func: Callable,
                          requires_weights: bool = False, requires_y_ensemble_pred: bool = False,
                          more_diversity_if_higher: bool = False, single_model_ensemble_default: float = 1):
    """Create an abstract usable diversity metric based on the input.

    Parameters
    ----------
    metric_type: str in {"OrdinalClassification", "NonOrdinalClassification", "Regression"}
        The type of diversity metric. Determines what type of input is passed to the metric function.
    metric_name: str
        Name of the metric.
    metric_func: Callable
        The function that is called to compute the diversity.
        We assume the function f to have a specific signature based on "metric_type":

            - "OrdinalClassification": f(y_pred_ensemble, Y_pred_base_models, y_true)

        With:
            - y_pred_ensemble: ndarray, (n_samples, ) or (n_samples, n_classes)
                Contains the (probability) predictions of the ensemble
            - Y_pred_base_models: List[ndarray], (n_samples, n_base_models) or (n_samples, n_classes, n_base_models)
                Contains the (probability) predictions of each base model
            - y_true: ndarray, (n_samples,)
                Ground truth / target vector
    requires_weights: bool, default=False
        If the metric requires the weights (for weighted average based metrics).
    requires_y_ensemble_pred: Bool, default=False
        If the metric requires the predictions from the ensemble.
    more_diversity_if_higher: bool, default=False
        If false, the diversity metrics represents more ensemble diversity with lower values. Otherwise, set this to
        true.
    single_model_ensemble_default: float, default=1
        What value to return if the ensemble consists only of a single model.


    Returns:
    -------
    Diversity Metric Depending on the Input

    """
    if metric_type == "OrdinalClassification":
        d_m = OrdinalClassificationDiversityMetric(metric_func, metric_name,
                                                   requires_weights, requires_y_ensemble_pred, more_diversity_if_higher,
                                                   single_model_ensemble_default)

    elif metric_type == "NonOrdinalClassification":
        d_m = NonOrdinalClassificationDiversityMetric(metric_func, metric_name,
                                                      requires_weights, requires_y_ensemble_pred,
                                                      more_diversity_if_higher, single_model_ensemble_default)
    else:
        raise NotImplementedError()

    return d_m


# - Error Correlation
def average_loss_correlation(y_true, Y_pred_base_models: list[np.ndarray], weights=None, aggregation_method=np.mean):
    """Loss correlation implementation. Compute the correlation of the loss for each instance w.r.t. the true class
    between all pairs of base models.

    """
    # TODO: add transformed labels as parameter in signature
    lb = LabelBinarizer()
    lb.fit(y_true)

    # Try to fix bad split situation
    # TODO: make this a global or model based fix somehow, pass classes?
    if len(lb.classes_) != Y_pred_base_models[0].shape[-1]:
        lb.classes_ = list(range(Y_pred_base_models[0].shape[-1]))

    transformed_labels = lb.transform(y_true)

    if transformed_labels.shape[1] == 1:
        transformed_labels = np.append(
            1 - transformed_labels, transformed_labels, axis=1,
        )

    loss_per_bm = [not_aggregated_loss(transformed_labels, bm_pred) for bm_pred in Y_pred_base_models]

    # Get pairwise correlations
    # following https://chrisalbon.com/code/machine_learning/feature_selection/drop_highly_correlated_features/
    corr_matrix = np.abs(np.corrcoef(loss_per_bm))

    # Set all values to nan except for the upper triangle of the matrix
    upper = corr_matrix[np.triu(np.ones(corr_matrix.shape), k=1).astype(bool)]

    # Replace nan values with 0 because it is equal to no correlation if something is just a constant predictor
    upper = np.nan_to_num(upper)

    return aggregation_method(upper)


def not_aggregated_loss(transformed_labels, y_pred, eps=1e-15):
    """Sklearn log loss adapted to stop early."""
    return 1 - (transformed_labels * y_pred).sum(axis=1)



# --- Initialization
LossCorrelation = make_diversity_metric("OrdinalClassification", "Average Loss Correlation",
                                        average_loss_correlation)



# Code Taken from here with adaptions to be usable:
# https://github.com/automl/auto-sklearn/blob/master/autosklearn/ensembles/ensemble_selection.py
from __future__ import annotations

import os
from collections import Counter
from collections.abc import Callable
from concurrent.futures import ProcessPoolExecutor

import numpy as np
from phem.base_utils.metrics import AbstractMetric
from phem.framework.abstract_weighted_ensemble import AbstractWeightedEnsemble
from sklearn.utils import check_random_state


class EnsembleSelection(AbstractWeightedEnsemble):
    """An ensemble of selected algorithms.

    Fitting an EnsembleSelection generates an ensemble from the models
    generated during the search process. Can be further used for prediction.

    Parameters
    ----------
    base_models: List[Callable], List[sklearn estimators]
        The pool of fitted base models.
    metric: AbstractMetric
        The metric used to evaluate the models
    n_jobs: int, default=-1
        Number of processes to use for parallelization. -1 means all available.
    random_state: Optional[int | RandomState] = None
        The random_state used for ensemble selection.
        *   None - Uses numpy's default RandomState object
        *   int - Successive calls to fit will produce the same results
        *   RandomState - Truely random, each call to fit will produce
                          different results, even with the same object.
    use_best: bool = True
        After finishing all iterations, use the best found ensemble instead of the last found ensemble.
    """

    def __init__(self, base_models: list[Callable], n_iterations: int, metric: AbstractMetric, n_jobs: int = -1,
                 random_state: int | np.random.RandomState | None = None,
                 use_best: bool = False) -> None:
        #base_models = base_models[:20]

        super().__init__(base_models, "predict_proba")
        self.ensemble_size = n_iterations
        self.metric = metric
        self.use_best = use_best

        # -- Code for multiprocessing
        if (n_jobs == 1) or (os.name == "nt"):
            self._use_mp = False
            if os.name == "nt":
                pass
        else:
            if n_jobs == -1:
                n_jobs = len(os.sched_getaffinity(0))
            self._n_jobs = n_jobs
            self._use_mp = True

        # Behaviour similar to sklearn
        #   int - Deteriministic with succesive calls to fit
        #   RandomState - Successive calls to fit will produce differences
        #   None - Uses numpmys global singleton RandomState
        # https://scikit-learn.org/stable/common_pitfalls.html#controlling-randomness
        self.random_state = random_state

    def ensemble_fit(self, predictions: list[np.ndarray], labels: np.ndarray) -> AbstractWeightedEnsemble:
        self.ensemble_size = int(self.ensemble_size)
        if self.ensemble_size < 1:
            raise ValueError("Ensemble size cannot be less than one!")
        if not isinstance(self.metric, AbstractMetric):
            raise ValueError("The provided metric must be an instance of a AbstractMetric, "
                             f"nevertheless it is {self.metric}({type(self.metric)})")

        self._fit(predictions, labels)
        self.apply_use_best()
        self._calculate_final_weights()

        # -- Set metadata correctly
        self.iteration_batch_size_ = len(predictions)

        return self

    def _fit(self, predictions: list[np.ndarray], labels: np.ndarray) -> None:
        """Fast version of Rich Caruana's ensemble selection method."""
        self.num_input_models_ = len(predictions)
        rand = check_random_state(self.random_state)

        ensemble = []  # type: List[np.ndarray]
        trajectory = []  # contains iteration best
        self.val_loss_over_iterations_ = []  # contains overall best
        order = []

        ensemble_size = self.ensemble_size

        weighted_ensemble_prediction = np.zeros(
            predictions[0].shape,
            dtype=np.float64,
        )
        fant_ensemble_prediction = np.zeros(
            weighted_ensemble_prediction.shape,
            dtype=np.float64,
        )

        for _i in range(ensemble_size):
            #print(i)
            s = len(ensemble)
            if s > 0:
                np.add(
                    weighted_ensemble_prediction,
                    ensemble[-1],
                    out=weighted_ensemble_prediction,
                )

            # -- Process Iteration Solutions
            if self._use_mp:
                losses = self._compute_losses_mp(weighted_ensemble_prediction, labels, predictions, s)
            else:
                losses = np.zeros(
                    (len(predictions)),
                    dtype=np.float64,
                )

                # Memory-efficient averaging!
                for j, pred in enumerate(predictions):
                    # fant_ensemble_prediction is the prediction of the current ensemble
                    # and should be ([predictions[selected_prev_iterations] + predictions[j])/(s+1)
                    # We overwrite the contents of fant_ensemble_prediction
                    # directly with weighted_ensemble_prediction + new_prediction and then scale for avg
                    np.add(
                        weighted_ensemble_prediction,
                        pred,
                        out=fant_ensemble_prediction,
                    )
                    np.multiply(
                        fant_ensemble_prediction,
                        (1. / float(s + 1)),
                        out=fant_ensemble_prediction,
                    )

                    losses[j] = self.metric(labels, fant_ensemble_prediction, to_loss=True)
            #print("LOSSES ",losses)

            # -- Eval Iteration results
            all_best = np.argwhere(losses == np.nanmin(losses)).flatten()

            best = rand.choice(all_best)
            ensemble_loss = losses[best]

            ensemble.append(predictions[best])
            #print("ENSEMBLE ", ensemble)
            trajectory.append(ensemble_loss)
            order.append(best)
            #print("order ", order)

            # Build Correct Validation loss list
            if not self.val_loss_over_iterations_ or self.val_loss_over_iterations_[-1] > ensemble_loss:
                self.val_loss_over_iterations_.append(ensemble_loss)
            else:
                self.val_loss_over_iterations_.append(self.val_loss_over_iterations_[-1])

            # -- Handle special cases
            if len(predictions) == 1:
                break

            # If we find a perfect ensemble/model, stop early
            if ensemble_loss == 0:
                break

        self.indices_ = order
        self.trajectory_ = trajectory
        #print("TRAJECTORY ", trajectory)

    def _compute_losses_mp(self, weighted_ensemble_prediction, labels, predictions, s):
        # -- Process Iteration Solutions
        func_args = (weighted_ensemble_prediction, labels, s, self.metric, predictions)
        pred_i_list = list(range(len(predictions)))

        with ProcessPoolExecutor(self._n_jobs, initializer=_pool_init, initargs=func_args) as ex:
            results = ex.map(_init_wrapper_evaluate_single_solution, pred_i_list)

        return np.array(list(results))

    def apply_use_best(self):
        if self.use_best:
            # Basically from autogluon the code
            min_score = np.min(self.trajectory_)
            idx_best = self.trajectory_.index(min_score)
            self.indices_ = self.indices_[:idx_best + 1]
            self.trajectory_ = self.trajectory_[:idx_best + 1]
            self.ensemble_size = idx_best + 1
            self.validation_loss_ = self.trajectory_[idx_best]
        else:
            self.validation_loss_ = self.trajectory_[-1]
            self.val_loss_over_iterations_ = self.trajectory_

    def _calculate_final_weights(self) -> None:
        ensemble_members = Counter(self.indices_).most_common()
        weights = np.zeros(
            (self.num_input_models_,),
            dtype=np.float64,
        )
        for ensemble_member in ensemble_members:
            weight = float(ensemble_member[1]) / self.ensemble_size
            weights[ensemble_member[0]] = weight

        if np.sum(weights) < 1:
            weights = weights / np.sum(weights)

        self.weights_ = weights


def _pool_init(_weighted_ensemble_prediction, _labels, _sample_size, _score_metric, _predictions):
    global p_weighted_ensemble_prediction, p_labels, p_sample_size, p_score_metric, p_predictions
    p_weighted_ensemble_prediction = _weighted_ensemble_prediction
    p_labels = _labels
    p_sample_size = _sample_size
    p_score_metric = _score_metric
    p_predictions = _predictions


def _init_wrapper_evaluate_single_solution(pred_index):
    return evaluate_single_solution(p_weighted_ensemble_prediction, p_labels, p_sample_size, p_score_metric,
                                    p_predictions[pred_index])


def evaluate_single_solution(weighted_ensemble_prediction, labels, sample_size, score_metric, pred):
    fant_ensemble_prediction = np.add(weighted_ensemble_prediction, pred)
    np.multiply(fant_ensemble_prediction, (1. / float(sample_size + 1)), out=fant_ensemble_prediction)

    return score_metric(labels, fant_ensemble_prediction, to_loss=True)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from math import isclose

import numpy as np
from phem.framework.abstract_ensemble import AbstractEnsemble


class AbstractWeightedEnsemble(AbstractEnsemble):
    __metaclass__ = ABCMeta

    def __init__(self, *args):
        super().__init__(*args)
        self.normalize_predict_proba_ = False
        self.model_specific_metadata_ = {}

    @staticmethod
    def _calculate_weights(discrete_weights) -> np.ndarray:
        return discrete_weights / np.sum(discrete_weights)


    @staticmethod
    def _calculate_counts(float_weights, sample_size) -> np.ndarray:
        return float_weights * sample_size


    @staticmethod
    def _ensemble_predict(predictions: list[np.ndarray], weights: np.ndarray,
                          normalize_predict_proba: bool = False) -> np.ndarray:
        """Blanket (not-the-most-efficient) ensemble predict for a weighted ensemble.

        Parameters
        ----------
        weights: np.ndarray
            Can be of any numeric range (but some metric might require normalization).
        normalize_predict_proba: bool, default=False
            If True, normalize the prediction probabilities such that they sum up to 1 and are in [0,1].
            Only needed if the weights are not in [0,1] or do not sum up to 1.
            We apply the softmax but only if negative weights are present.
            We apply simple normalization if weights are positive but do not sum to 1.
        """
        average = np.zeros_like(predictions[0], dtype=np.float64)
        tmp_predictions = np.empty_like(predictions[0], dtype=np.float64)

        for pred, weight in zip(predictions, weights, strict=False):
            np.multiply(pred, weight, out=tmp_predictions)
            np.add(average, tmp_predictions, out=average)

        if normalize_predict_proba:
            if any(weights < 0):
                exp = np.nan_to_num(np.exp(np.clip(average, -88.72, 88.72)))  # Clip to avoid overflow
                average = exp / exp.sum(axis=1)[:, None]
                average = average / average.sum(axis=1)[:, None]
            elif not isclose(weights.sum(), 1):
                average = average / average.sum(axis=1)[:, None]

        return average

    @abstractmethod
    def ensemble_fit(self, base_models_predictions: list[np.ndarray], labels: np.ndarray) -> AbstractEnsemble:
        pass

    @property
    def attr_after_fit(self):
        return ["weights_", "validation_loss_", "iteration_batch_size_", "val_loss_over_iterations_"]

    @property
    def check_fitted(self):

        return [req_attr for req_attr in self.attr_after_fit if not hasattr(self, req_attr)]

    def ensemble_predict(self, predictions: np.ndarray | list[np.ndarray]) -> np.ndarray:
        if self.check_fitted:
            raise ValueError("Model not fitted! Missing fit parameters are:", self.check_fitted)

        return self._confidences_to_predictions(self.ensemble_predict_proba(predictions))

    def ensemble_predict_proba(self, predictions: np.ndarray | list[np.ndarray]) -> np.ndarray:
        if len(predictions) == len(self.weights_):
            # predictions include those of zero-weight models.
            return self._ensemble_predict(predictions, self.weights_,
                                          normalize_predict_proba=self.normalize_predict_proba_)

        elif len(predictions) == np.count_nonzero(self.weights_):
            # predictions do not include those of zero-weight models.
            non_null_weights = self.weights_[self.weights_ != 0]
            return self._ensemble_predict(predictions, non_null_weights,
                                          normalize_predict_proba=self.normalize_predict_proba_)

        # If none of the above applies, then something must have gone wrong.
        else:
            raise ValueError("The dimensions of ensemble predictions"
                             " and ensemble weights do not match!")

    @property
    def _to_save_metadata(self):
        """Standard metadata for weighted ensemble.

        Parameters
        -------
        self.weights_: np.ndarray
            The best found weights
        self.validation_loss_: float
            The validation loss of the best found weights
        self.iteration_batch_size_: int
            The batch size used in each iteration of the selection method.
            This is equal to the number of evolutions that are done inbetween checking whether
            we found a new best weight vector.
        self.val_loss_over_iterations_: List[float]
            The best validation loss over each iteration of the search.
        """
        # Need to transform any numpy datatypes to normal python dtypes such that we can use JSON
        metadata = {
            "weights_": self.weights_.tolist(),
            "validation_loss_": float(self.validation_loss_),
            "iteration_batch_size_": int(self.iteration_batch_size_),
            "val_loss_over_iterations_": np.array(self.val_loss_over_iterations_).tolist(),
        }
        metadata.update(self.model_specific_metadata_)

        return metadata

    # -- Experimental Stuff
    def plot_opt_history(self):
        # print(self.model_specific_metadata_)

        import matplotlib.pyplot as plt
        import pandas as pd

        t = pd.Series(self.val_loss_over_iterations_)
        t.plot()
        plt.show()



from __future__ import annotations

from phem.methods.ensemble_selection.qdo.behavior_spaces import (
    get_bs_configspace_similarity_and_loss_correlation,
    get_bs_ensemble_size_and_loss_correlation,
)
from phem.methods.ensemble_selection.qdo.qdo_es import QDOEnsembleSelection

__all__ = [
    "QDOEnsembleSelection", "get_bs_configspace_similarity_and_loss_correlation", "get_bs_ensemble_size_and_loss_correlation",
]



"""Code to easily call the post hoc ensembling methods."""
from __future__ import annotations

from phem.application_utils import methods_factories


def get_ensemble_switch_case_config(method, rng_seed=None, metric=None, n_jobs=None,
                                    is_binary=None, labels=None):


    if method == "SingleBest":
        return methods_factories._factory_SingleBest(metric=metric)
    # -- Best Configs on Val data from Purucker et al. QDO-ES Paper
    elif method == "GES":
        return methods_factories._factory_es(rng_seed, metric, n_jobs, True)
    elif method == "QDO-ES":
        return methods_factories._factory_qdo(
            rng_seed, metric, is_binary, labels, n_jobs, "sliding",
            "bs_configspace_similarity_and_loss_correlation", 49, "RandomL2Combinations",
            1, "combined_dynamic", "two_point_crossover", 0.5,
            True, 0.5, True, None,None, None, 1., None, 40,
        )
    elif method == "QO-ES":
        return methods_factories._factory_qdo(
            rng_seed, metric, is_binary, labels, n_jobs, "quality",
            None, 49, "RandomL2Combinations",
            1, "combined_dynamic", "average", 0.5,
            True, 0.5, True, None,None, None, 1., None, 40,
        )
    else:
        raise ValueError(f"Unknown method! Got: {method}")



from __future__ import annotations

import numpy as np
from ribs.archives._archive_base import ArchiveBase


def get_quantile_mask(values_to_mask, quantile):
    """Remarks:
    We assume higher is better for the value. Thus, 0.75 -> gives us the first quantile w.r.t. better.
    """
    min_perf_val = np.quantile(values_to_mask, quantile)
    return values_to_mask >= min_perf_val


def remap_clean_up(self, reset_bounds=True):
    """Remap clean up code:

    * Reset the upper and lower bound
    * re-adds all solutions (From buffer and last elites). The last added solution will be the solution with which
        add() was called.

    """
    if reset_bounds:
        self._lower_bounds = np.array(
            [bound[0] for bound in self._boundaries])
        self._upper_bounds = np.array([
            bound[dim] for bound, dim in zip(self._boundaries, self._dims, strict=False)
        ])

    old_sols = self._solutions[self._occupied_indices_cols].copy()
    old_objs = self._objective_values[self._occupied_indices_cols].copy()
    old_behs = self._behavior_values[self._occupied_indices_cols].copy()
    old_metas = self._metadata[self._occupied_indices_cols].copy()

    self.clear()
    for sol, obj, beh, meta in zip(old_sols, old_objs, old_behs, old_metas, strict=False):
        # Add solutions from old archive.
        status, value = ArchiveBase.add(self, sol, obj, beh, meta)
    for sol, obj, beh, meta in self._buffer:
        # Add solutions from buffer.
        status, value = ArchiveBase.add(self, sol, obj, beh, meta)

    if not self._first_remap_done:
        self._first_remap_done = True
        self._remap_frequency = self._default_remap_frequency

    return status, value



# Potentially useful metrics for evaluation wrapped in an easier to use object
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from typing import TYPE_CHECKING

import numpy as np
from sklearn.utils.validation import _check_y, check_array

if TYPE_CHECKING:
    import pandas as pd


# -- Metric Utils
def make_metric(metric_func: Callable, metric_name: str, maximize: bool,
                classification: bool, always_transform_conf_to_pred: bool,
                optimum_value: int, pos_label: int = 1, requires_confidences: bool = False,
                only_positive_class: bool = False):
    """Make a metric that has additional information.

    Parameters
    ----------
    metric_func: Callable
        The metric function to call.
        We expect it to be metric_func(y_true, y_pred) with y_pred potentially being
        probabilities instead of classes.
    metric_name: str
        Name of the metric
    maximize: bool
        Whether to maximize the metric or not
    classification: bool
        If it is a classification metric or not
    always_transform_conf_to_pred: bool
        Set to Ture if the metric can not handle confidences and only accepts predictions (only for classification)
    optimum_value: int
        The maximal value the metric can reach (used to compute the loss).
    pos_label: int, default=1
        Index of the label used as positive label (relevant only for binary classification metrics)
    requires_confidences: bool, default=False
        If the metric requires confidences.
    only_positive_class: bool, default=False
        Only relevant if requires_confidences is True. If only_positive_class is true, only the positive class
        values are passed. This is only needed for binary classification. Ignored if always_transform_conf_to_pred is
        True.
    """
    return AbstractMetric(metric_func, metric_name, maximize, classification,
                          always_transform_conf_to_pred, optimum_value, pos_label,
                          requires_confidences, only_positive_class)


class AbstractMetric:
    """Abstract Metric used in some codes.

    We transform confidences to prediction if needed for a metric and the model does it not by itself.
    Thereby, we assume y_true to be integers because we transform y_pred into integers as well.

    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, metric, name, maximize, classification, transform_conf_to_pred, optimum_value, pos_label,
                 requires_confidences, only_positive_class):
        self.metric = metric
        self.maximize = maximize
        self.name = name
        self.classification = classification
        self.transform_conf_to_pred = transform_conf_to_pred
        self.optimum_value = optimum_value
        self.pos_label = pos_label
        self.threshold = 0.5
        self.requires_confidences = requires_confidences
        self.only_positive_class = only_positive_class

    def __call__(self, y_true: pd.DataFrame | np.ndarray, y_pred: pd.DataFrame | np.ndarray,
                 to_loss: bool = False, checks=True):
        """Parameters
        ----------
        y_true: array-like
            ground truth, assumed to be integers!
        y_pred: array-like
            Either confidences/probabilities matrix (n_samples, n_classes) or prediction vector (n_samples, )
            If not classification, only prediction vector is allowed for now.
            If confidences, we expect the order of n_classes to be identical to the order of np.unique(y_true).
        to_loss: bool
            Whether to return the loss or not
        """
        # -- Input validation
        if checks:
            y_true = _check_y(y_true)

        if not self.classification:
            if checks:
                y_pred = _check_y(y_pred, y_numeric=True)
        elif y_pred.ndim == 1:
            if checks:
                y_pred = _check_y(y_pred)

                if self.requires_confidences:
                    raise ValueError("Confidences are needed for this metric but predictions are passed.")
        elif y_pred.ndim == 2:
            if checks:
                y_pred = check_array(y_pred)

            # - Special case if metric can not handle confidences
            if self.transform_conf_to_pred:
                y_pred = np.argmax(y_pred, axis=1)
            elif self.only_positive_class:
                y_pred = y_pred[:, self.pos_label]

        else:
            raise ValueError(f"y_pred has to many dimensions! Found ndim: {y_pred.ndim}")

        # --- Call metric
        metric_value = self.metric(y_true, y_pred)

        # --- Return
        if to_loss:
            return self.to_loss(metric_value)

        return metric_value

    def to_loss(self, metric_value):
        # General Purpose Loss: the absolute difference to the optimum
        #   -> smaller is always better
        return abs(self.optimum_value - metric_value)

    def inverse_loss(self, loss_value):

        if (self.optimum_value == 0) and (not self.maximize):
            return loss_value

        # FIXME: this ignores negative metric_values or optima
        return self.optimum_value - loss_value



from __future__ import annotations

from sklearn.compose import ColumnTransformer, make_column_selector
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OrdinalEncoder


def get_default_preprocessing():
    # Preprocessing only for ensemble methods that use the original data.
    # As no of our methods uses the original data but only the predictions, this is technically not used and only
    # here for compatibility.
    return ColumnTransformer(
        transformers=[
            ("num", SimpleImputer(strategy="constant", fill_value=-1),
             make_column_selector(dtype_exclude="category")),
            ("cat", Pipeline(steps=[("encoder", OrdinalEncoder(handle_unknown="use_encoded_value", unknown_value=-1)),
                                    ("imputer", SimpleImputer(strategy="constant", fill_value=-1)),
                                    ]),
             make_column_selector(dtype_include="category")),
        ],
        sparse_threshold=0,
    )



"""Tool to more easily get and support different metrics."""
from __future__ import annotations

from functools import partial

from sklearn.metrics import balanced_accuracy_score

from phem.base_utils.custom_metrics.roc_auc import roc_auc_score
from phem.base_utils.metrics import make_metric


def msc(metric_name, is_binary, labels):
    if metric_name == "balanced_accuracy":
        return make_metric(balanced_accuracy_score,
                           metric_name="balanced_accuracy",
                           maximize=True,
                           classification=True,
                           always_transform_conf_to_pred=True,
                           optimum_value=1,
                           requires_confidences=False)

    elif metric_name == "roc_auc":
        if is_binary:
            return make_metric(roc_auc_score,
                               metric_name="roc_auc",
                               maximize=True,
                               classification=True,
                               always_transform_conf_to_pred=False,
                               optimum_value=1,
                               requires_confidences=True,
                               only_positive_class=True,
                               pos_label=1)
        else:
            return make_metric(partial(roc_auc_score, average="macro", multi_class="ovr", labels=labels),
                               metric_name="roc_auc",
                               maximize=True,
                               classification=True,
                               always_transform_conf_to_pred=False,
                               optimum_value=1,
                               requires_confidences=True)
    else:
        raise ValueError(f"Unknown metric name: {metric_name}")



"""Implicit Diversity Metrics.

Any metric that tries to implicit measure the diversity of an ensemble.
"""
from __future__ import annotations

from itertools import combinations
from statistics import mean

from phem.methods.ensemble_selection.qdo.behavior_space import BehaviorFunction


def config_space_gower_similarity(input_metadata: tuple[dict[str, int | float], list[dict]]):
    """Gower Distance used to compute the similarity between the identical keys of all pairs of configurations.

    The input metadata consists of one directory contain the ranges for numeric-valued keys of configurations and
        the configurations of the base models in the ensemble.

    The gower distance is transformed into similarity. This allows us the following interpretation:
        1 - No Diversity
        0 - Maximal Diversity
    """
    ranges, base_models_metadata = input_metadata

    # Default Case
    if len(base_models_metadata) == 1:
        return 1

    # Iterate over all pairs
    scores_over_pairs = []
    for bm_md_1, bm_md_2 in combinations(base_models_metadata, 2):
        identical_keys = set(bm_md_1.keys()).intersection(set(bm_md_2.keys()))
        distance_scores = []  # lower is better

        for i_k in identical_keys:

            if isinstance(bm_md_1[i_k], str):
                # -- Assume it is a categorical hyperparameter/choice
                if bm_md_1[i_k] == bm_md_2[i_k]:
                    distance_scores.append(0)  # Identical means absolute similar / close
                else:
                    # Unequal means punishing by maximal distance (no config space interpretation added)
                    distance_scores.append(1)
            elif bm_md_1[i_k] == bm_md_2[i_k]:
                # Catch edge case where the values are identical to avoid potential divide by zero error if
                #   bm_md_1 and bm_md_2 equal the only existing unique value.
                distance_scores.append(0)
            else:
                distance_scores.append(abs(bm_md_1[i_k] - bm_md_2[i_k]) / ranges[i_k])

        sim_scores = [1 - v for v in distance_scores]
        scores_over_pairs.append(mean(sim_scores))

    return mean(scores_over_pairs)


ConfigSpaceGowerSimilarity = BehaviorFunction(config_space_gower_similarity, ["input_metadata"], (0, 1), "none",
                                              name="Gower Similarity in Config Space (Lower is more Diverse)")



"""Factories to call the post hoc ensembling methods (with a given configuration)."""
from __future__ import annotations

def _factory_SingleBest(metric=None, **kwargs):
    # "custom.SingleBest"
    from phem.methods.baselines.single_best import SingleBest

    return {
        "technique": SingleBest,
        "technique_args": {"metric": metric,
                           "predict_method": "predict_proba"},
        "pre_fit_base_models": True
    }

def _factory_es(rng_seed, metric, n_jobs, use_best):
    from numpy.random import RandomState

    from phem.methods.ensemble_selection.greedy_ensemble_selection import EnsembleSelection

    return {
        "technique": EnsembleSelection,
        "technique_args": {"n_iterations": 50,
                           "metric": metric,
                           "n_jobs": n_jobs,
                           "random_state": RandomState(rng_seed),
                           "use_best": use_best},
        "pre_fit_base_models": True,
    }

def _factory_qdo(rng_seed, metric, is_binary, labels, n_jobs, archive_type, behavior_space_choice, max_elites,
                 emitter_initialization_method, starting_step_size, elite_selection_method, crossover_choice,
                 crossover_probability, crossover_probability_dynamic, mutation_probability_after_crossover,
                 mutation_probability_after_crossover_dynamic, negative_steps, weight_random_elite_selection,
                 weight_random_step_selection, buffer_ratio_choice, dynamic_updates_consider_rejections,
                 batch_size):
    from numpy.random import RandomState

    from phem.methods.ensemble_selection.qdo.qdo_es import QDOEnsembleSelection

    # -- Get Parameter
    if behavior_space_choice is None:
        bs = None
    elif behavior_space_choice == "bs_configspace_similarity_and_loss_correlation":
        from phem.methods.ensemble_selection.qdo.behavior_spaces import get_bs_configspace_similarity_and_loss_correlation
        bs = get_bs_configspace_similarity_and_loss_correlation()
    else:
        raise ValueError("Unknown choice for behavior_space!")

    buffer_ratio = buffer_ratio_choice if buffer_ratio_choice is not None else 1.

    # - Emitter Vars
    emitter_vars = {
        "starting_step_size": starting_step_size,
        "elite_selection_method": elite_selection_method,
    }

    if crossover_choice == "no_crossover":
        emitter_vars["crossover"] = None
    else:
        emitter_vars["crossover"] = crossover_choice

        if isinstance(crossover_probability, float):
            emitter_vars["crossover_probability"] = crossover_probability
        else:
            raise ValueError("crossover_probability must be a float!")

        if isinstance(mutation_probability_after_crossover, float):
            emitter_vars["mutation_probability_after_crossover"] = mutation_probability_after_crossover
        else:
            raise ValueError("mutation_probability_after_crossover must be a float!")

        if crossover_probability_dynamic is not None:
            emitter_vars["crossover_probability_dynamic"] = crossover_probability_dynamic
        if mutation_probability_after_crossover_dynamic is not None:
            emitter_vars["mutation_probability_after_crossover_dynamic"] = mutation_probability_after_crossover_dynamic

    if negative_steps is not None:
        emitter_vars["negative_steps"] = negative_steps

    if weight_random_elite_selection is not None:
        emitter_vars["weight_random_elite_selection"] = weight_random_elite_selection

    if weight_random_step_selection is not None:
        emitter_vars["weight_random_step_selection"] = weight_random_step_selection

    if dynamic_updates_consider_rejections is not None:
        emitter_vars["dynamic_updates_consider_rejections"] = dynamic_updates_consider_rejections

    return {
        "technique": QDOEnsembleSelection,
        "technique_args": {
            "n_iterations": 50,
            "batch_size": batch_size,
            "score_metric": metric,
            "max_elites": max_elites,
            "archive_type": archive_type,
            "buffer_ratio": buffer_ratio,
            "behavior_space": bs,
            "emitter_initialization_method": emitter_initialization_method,
            "emitter_vars": emitter_vars,
            "random_state": RandomState(rng_seed),
            "n_jobs": n_jobs,
        },
        "pre_fit_base_models": True,
    }



"""Contains the SlidingBoundariesArchive."""
from __future__ import annotations

from collections import deque

import numba as nb
import numpy as np
from ribs.archives._archive_base import ArchiveBase, require_init
from sortedcontainers import SortedList

from phem.methods.ensemble_selection.qdo.custom_archives.utils import remap_clean_up

_EPSILON = 1e-6


class SolutionBuffer:
    """An internal class that stores relevant data to re-add after remapping.

    Maintains two data structures:
    - Queue storing the buffer_capacity last entries (solution + objective value
      + behavior values + metadata). When new items are inserted, the oldest
      ones are popped.
    - Sorted lists with the sorted behavior values in each dimension. Behavior
      values are removed from theses lists when they are removed from the queue.
    """

    def __init__(self, buffer_capacity, behavior_dim):
        self._buffer_capacity = buffer_capacity
        self._queue = deque()
        self._bc_lists = [SortedList() for _ in range(behavior_dim)]
        self._iter_idx = 0

    def __iter__(self):
        """Enables iterating over solutions stored in the buffer."""
        return self

    def __next__(self):
        """Returns the next entry in the buffer."""
        if self._iter_idx >= self.size:
            self._iter_idx = 0
            raise StopIteration
        result = self._queue[self._iter_idx]
        self._iter_idx += 1
        return result

    def add(self, solution, objective_value, behavior_values, metadata=None):
        """Inserts a new entry.

        Pops the oldest if it is full.
        """
        if self.full():
            # Remove item from the deque.
            _, _, bc_deleted, _ = self._queue.popleft()
            # Remove bc from sorted lists.
            for i, bc in enumerate(bc_deleted):
                self._bc_lists[i].remove(bc)

        self._queue.append(
            (solution, objective_value, behavior_values, metadata))

        # Add bc to sorted lists.
        for i, bc in enumerate(behavior_values):
            self._bc_lists[i].add(bc)

    def full(self):
        """Whether buffer is full."""
        return len(self._queue) >= self._buffer_capacity

    @property
    def sorted_behavior_values(self):
        """(behavior_dim, self.size) numpy.ndarray: Sorted behavior values of
        each dimension.
        """
        return np.array(self._bc_lists, dtype=np.float64)

    @property
    def size(self):
        """Number of solutions stored in the buffer."""
        return len(self._queue)

    @property
    def capacity(self):
        """Capacity of the buffer."""
        return self._buffer_capacity


class SlidingBoundariesArchive(ArchiveBase):
    """An archive with a fixed number of sliding boundaries on each dimension.

    This archive is the container described in `Fontaine 2019
    <https://arxiv.org/abs/1904.10656>`_. Just like the
    :class:`~ribs.archives.GridArchive`, it can be visualized as an
    n-dimensional grid in the behavior space that is divided into a certain
    number of bins in each dimension. Internally, this archive stores a buffer
    with the ``buffer_capacity`` most recent solutions and uses them to
    determine the boundaries of the behavior characteristics along each
    dimension. After every ``remap_frequency`` solutions are inserted, the
    archive remaps the boundaries based on the solutions in the buffer.

    Initially, the archive has no solutions, so it cannot automatically
    calculate the boundaries. Thus, until the first remap, this archive divides
    the behavior space defined by ``ranges`` into equally sized bins.

    Overall, this archive attempts to make the distribution of the space
    illuminated by the archive more accurately match the true distribution of
    the behavior characteristics when they are not uniformly distributed.

    Args:
        dims (array-like): Number of bins in each dimension of the behavior
            space, e.g. ``[20, 30, 40]`` indicates there should be 3 dimensions
            with 20, 30, and 40 bins. (The number of dimensions is implicitly
            defined in the length of this argument).
        ranges (array-like of (float, float)): `Initial` upper and lower bound
            of each dimension of the behavior space, e.g. ``[(-1, 1), (-2, 2)]``
            indicates the first dimension should have bounds :math:`[-1,1]`
            (inclusive), and the second dimension should have bounds
            :math:`[-2,2]` (inclusive). ``ranges`` should be the same length as
            ``dims``.
        seed (int): Value to seed the random number generator. Set to None to
            avoid a fixed seed.
        dtype (str or data-type): Data type of the solutions, objective values,
            and behavior values. We only support ``"f"`` / :class:`np.float32`
            and ``"d"`` / :class:`np.float64`.
        remap_frequency (int): Frequency of remapping. Archive will remap once
            after ``remap_frequency`` number of solutions has been found.
        buffer_capacity (int): Number of solutions to keep in the buffer.
            Solutions in the buffer will be reinserted into the archive after
            remapping.
    """

    def __init__(self,
                 dims,
                 ranges,
                 seed=None,
                 dtype=np.float64,
                 remap_frequency=100,
                 initial_remap=100,
                 buffer_capacity=1000,
                 show_analysis=False):
        self._dims = np.array(dims)
        if len(self._dims) != len(ranges):
            raise ValueError(f"dims (length {len(self._dims)}) and ranges "
                             f"(length {len(ranges)}) must be the same length")

        ArchiveBase.__init__(
            self,
            storage_dims=tuple(self._dims),
            behavior_dim=len(self._dims),
            seed=seed,
            dtype=dtype,
        )

        ranges = list(zip(*ranges, strict=False))
        self._lower_bounds = np.array(ranges[0], dtype=self.dtype)
        self._upper_bounds = np.array(ranges[1], dtype=self.dtype)
        self._interval_size = self._upper_bounds - self._lower_bounds

        # Specifics for sliding boundaries.
        self._default_remap_frequency = remap_frequency
        self._remap_frequency = initial_remap
        self._first_remap_done = False

        # Allocate an extra entry in each row so we can put the upper bound at
        # the end.
        self._boundaries = np.full((self._behavior_dim, np.max(self._dims) + 1),
                                   np.nan,
                                   dtype=self.dtype)

        # Initialize the boundaries.
        for i, (dim, lower_bound, upper_bound) in enumerate(
                zip(self._dims, self._lower_bounds, self._upper_bounds, strict=False)):
            self._boundaries[i, :dim + 1] = np.linspace(lower_bound,
                                                        upper_bound, dim + 1)
        self._boundaries_over_time = [self._boundaries.copy()]

        # Create buffer.
        self._buffer = SolutionBuffer(buffer_capacity, self._behavior_dim)

        # Total number of solutions encountered.
        self._total_num_sol = 0

        self._boundaries_over_time = []
        self._elites_over_time = []
        self._show_analysis = show_analysis

    @property
    def dims(self):
        """(behavior_dim,) numpy.ndarray: Number of bins in each dimension."""
        return self._dims

    @property
    def lower_bounds(self):
        """(behavior_dim,) numpy.ndarray: Lower bound of each dimension."""
        return self._lower_bounds

    @property
    def upper_bounds(self):
        """(behavior_dim,) numpy.ndarray: Upper bound of each dimension."""
        return self._upper_bounds

    @property
    def interval_size(self):
        """(behavior_dim,) numpy.ndarray: The size of each dim (upper_bounds -
        lower_bounds).
        """
        return self._interval_size

    @property
    def remap_frequency(self):
        """int: Frequency of remapping. Archive will remap once after
        ``remap_frequency`` number of solutions has been found.
        """
        return self._remap_frequency

    @property
    def buffer_capacity(self):
        """int: Maximum capacity of the buffer."""
        return self._buffer.capacity

    @property
    def boundaries(self):
        """List of numpy.ndarray: The dynamic boundaries of the bins in each
        dimension.

        Entry ``i`` in this list is an array that contains the boundaries of the
        bins in dimension ``i``. The array contains ``self.dims[i] + 1`` entries
        laid out like this::

            Archive bins:   | 0 | 1 |   ...   |    self.dims[i]    |
            boundaries[i]:  0   1   2   self.dims[i] - 1     self.dims[i]

        Thus, ``boundaries[i][j]`` and ``boundaries[i][j + 1]`` are the lower
        and upper bounds of bin ``j`` in dimension ``i``. To access the lower
        bounds of all the bins in dimension ``i``, use ``boundaries[i][:-1]``,
        and to access all the upper bounds, use ``boundaries[i][1:]``.
        """
        return [
            bound[:dim + 1] for bound, dim in zip(self._boundaries, self._dims, strict=False)
        ]

    @staticmethod
    @nb.jit(nopython=True)
    def _get_index_numba(behavior_values, upper_bounds, lower_bounds,
                         boundaries, dims):
        """Numba helper for get_index().

        See get_index() for usage.
        """
        behavior_values = np.minimum(
            np.maximum(behavior_values + _EPSILON, lower_bounds),
            upper_bounds - _EPSILON)
        index = []
        for i, behavior_value in enumerate(behavior_values):
            idx = np.searchsorted(boundaries[i][:dims[i]], behavior_value)
            index.append(max(0, idx - 1))
        return index

    def get_index(self, behavior_values):
        """Returns indices of the behavior values within the archive's grid.

        First, values are clipped to the bounds of the behavior space. Then, the
        values are mapped to bins via a binary search along the boundaries in
        each dimension.

        The indices can be used to access boundaries of a behavior value's bin.
        For example, the following retrieves the lower and upper bounds of the
        bin along dimension 0::

            idx = archive.get_index(...)  # Other methods also return indices.
            lower = archive.boundaries[0][idx[0]]
            upper = archive.boundaries[0][idx[0] + 1]

        See :attr:`boundaries` for more info.

        Args:
            behavior_values (numpy.ndarray): (:attr:`behavior_dim`,) array of
                coordinates in behavior space.

        Returns:
            tuple of int: The grid indices.
        """
        index = SlidingBoundariesArchive._get_index_numba(
            behavior_values, self.upper_bounds, self.lower_bounds,
            self._boundaries, self._dims)
        return tuple(index)

    @staticmethod
    @nb.jit(nopython=True)
    def _remap_numba_helper(sorted_bc, buffer_size, boundaries, behavior_dim,
                            dims):
        """Numba helper for _remap().

        See _remap() for usage.
        """
        for i in range(behavior_dim):
            for j in range(dims[i]):
                sample_idx = int(j * buffer_size / dims[i])
                boundaries[i][j] = sorted_bc[i][sample_idx]
            # Set the upper bound to be the greatest BC.
            boundaries[i][dims[i]] = sorted_bc[i][-1]

    def _remap(self):
        """Remaps the archive.

        The boundaries are relocated to the percentage marks of the distribution
        of solutions stored in the archive.

        Also re-adds all of the solutions to the archive.

        Returns:
            tuple: The result of calling :meth:`ArchiveBase.add` on the last
            item in the buffer.
        """
        # Sort all behavior values along the axis of each bc.
        sorted_bc = self._buffer.sorted_behavior_values

        # Calculate new boundaries.
        SlidingBoundariesArchive._remap_numba_helper(sorted_bc,
                                                     self._buffer.size,
                                                     self._boundaries,
                                                     self._behavior_dim,
                                                     self.dims)

        return remap_clean_up(self)

    @require_init
    def add(self, solution, objective_value, behavior_values, metadata=None):
        """Attempts to insert a new solution into the archive.

        This method remaps the archive after every :attr:`remap_frequency`
        solutions are added. Remapping involves changing the boundaries of the
        archive to the percentage marks of the behavior values stored in the
        buffer and re-adding all of the solutions stored in the buffer `and` the
        current archive.

        Args:
            solution (array-like): See :meth:`ArchiveBase.add`
            objective_value (float): See :meth:`ArchiveBase.add`
            behavior_values (array-like): See :meth:`ArchiveBase.add`
            behavior_values (object): See :meth:`ArchiveBase.add`
        Returns:
            See :meth:`ArchiveBase.add`
        """
        solution = np.asarray(solution)
        behavior_values = np.asarray(behavior_values)

        self._buffer.add(solution, objective_value, behavior_values, metadata)
        self._total_num_sol += 1

        if self._total_num_sol % self._remap_frequency == 0:
            if self._show_analysis and (not self._first_remap_done):
                self._boundaries_over_time.append(self._boundaries.copy())
                self._elites_over_time.append([(e.obj, *e.beh) for e in self])

            status, value = self._remap()

            if self._show_analysis:
                self._boundaries_over_time.append(self._boundaries.copy())
                self._elites_over_time.append([(e.obj, *e.beh) for e in self])
        else:
            status, value = ArchiveBase.add(self, solution, objective_value,
                                            behavior_values, metadata)
        return status, value



from __future__ import annotations

import numpy as np
from sklearn.utils.validation import check_is_fitted

from phem.base_utils.metrics import AbstractMetric
from phem.framework.abstract_ensemble import AbstractEnsemble


class SingleBest(AbstractEnsemble):
    """Single Best Selector.

    Parameters
    ----------
    base_models: List[Callable], List[sklearn estimators]
        The pool of fitted base models.
    metric : AbstractMetric function, default=None
        The metric function that should be used to determine the single best algorithm
        Special format required due to OpenML's metrics and our usage.
    predict_method: {"predict", "predict_proba"}, default="predict"
        If "predict" is selected, we determine the SB by passing the raw predictions to the metric.
        If "predict_proba" is selected, we determine the SB by passing the confidences to the metric.
        If the metric can not handle confidences and "predict_proba" is selected, the behavior is identical to
            when "predict" would have been selected.

    Attributes:
    ----------
    best_model_index_ : int
        The Index of Single Best Model found during :meth:`ensemble_fit`.
    selected_indices_: ndarray, shape (n_samples,)
        The selected indices per samples found during :meth:`ensemble_predict`. Used to determine selection performance.
    """

    def __init__(self, base_models, metric: AbstractMetric, predict_method="predict"):
        super().__init__(base_models, predict_method, predict_method_ensemble_predict="dynamic")
        self.predict_method = predict_method
        self.metric = metric
        self.predict_proba_input = "predict_proba"

    def ensemble_fit(self, base_models_predictions: list[np.ndarray], labels: np.ndarray) -> AbstractEnsemble:
        """Find the single best algorithm and store it for later."""
        if not isinstance(self.metric, AbstractMetric):
            raise ValueError("The provided metric must be an instance of a AbstractMetric, "
                             f"nevertheless it is {self.metric}({type(self.metric)})")

        performances = [self.metric(labels, bm_prediction, to_loss=True) for bm_prediction in base_models_predictions]
        self.best_model_index_ = np.argmin(performances)
        self.validation_loss_ = np.min(performances)

        return self

    def ensemble_predict(self, base_models_predictions: list[np.ndarray]) -> np.ndarray:
        """Return the predictions of the Single Best."""
        check_is_fitted(self, ["best_model_index_"])

        n_samples = base_models_predictions[0].shape[0]
        self.selected_indices_ = np.full(n_samples, self.best_model_index_)

        return np.array(base_models_predictions)[self.selected_indices_, np.arange(n_samples)]

    def ensemble_predict_proba(self, base_models_predictions: list[np.ndarray]) -> np.ndarray:
        """Return the predictions of the Single Best."""
        check_is_fitted(self, ["best_model_index_"])

        n_samples = base_models_predictions[0].shape[0]
        self.selected_indices_ = np.full(n_samples, self.best_model_index_)

        return base_models_predictions[self.best_model_index_]

    @property
    def _to_save_metadata(self):
        return {
            "best_model_index_": int(self.best_model_index_),
            "validation_loss_": float(self.validation_loss_),
        }



# Code Taken from here with (heavy) adaptions to be usable:
# https://github.com/automl/auto-sklearn/blob/master/autosklearn/ensembles/abstract_ensemble.py
from __future__ import annotations

from abc import ABCMeta, abstractmethod

import numpy as np
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y


class AbstractEnsemble:
    """Abstract class to guarantee that we can use autosklearn's ensemble techniques and custom techniques.

    During fit (for classification), we transform all labels into integers. This also happens for the predictions of
    base models.

    Parameters
    ----------
    base_models: List[Callable], List[sklearn estimators]
        The pool of fitted base models.
    predict_method: {"predict", "predict_proba"}, default="predict"
        Determine the predict method that is used to obtain the output of base models that is passed to an ensemble's
        fit method.
    predict_method_ensemble_predict: {"predict", "predict_proba", "dynamic", None}, default=None
        Determine the predict method that is used to obtain the output of base models that is passed to an ensemble's
        predict method.
        If None, the same method passed to ensemble fit is passed to ensemble predict.
        If "dynamic", the method is selected based on the predict or predict_proba call.
    passthrough : bool, default=False
        When False, only the predictions or confidences of the base models will be passed to the ensemble fit and
        predict. When True, the original training data is passed additionally. The ensemble technique must support this!

    Attributes:
    ----------
    le_ : LabelEncoder, object
        The label encoder created at :meth:`fit`.
    classes_ : ndarray, shape (n_classes,)
        The classes seen at :meth:`fit`.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, base_models, predict_method: str = "predict",
                 predict_method_ensemble_predict: str | None = None, passthrough: bool = False):

        self.base_models = base_models
        self.predict_method = predict_method
        self.passthrough = passthrough

        if predict_method_ensemble_predict is None:
            self.predict_method_ensemble_predict = predict_method
        else:

            if predict_method_ensemble_predict not in ["predict", "predict_proba", "dynamic"]:
                raise ValueError("The value for predict_method_ensemble_predict is not allowed.")

            self.predict_method_ensemble_predict = predict_method_ensemble_predict

        # -- Obtain metadata in a standardized format
        self.base_models_metadata_exists = False
        if any(hasattr(bm, "model_metadata") and (bm.model_metadata is not None) for bm in self.base_models):
            # The metadat will be stored in the same order as the predictions base_models_predictions
            self.base_models_metadata_ = [bm.model_metadata for bm in self.base_models]
            self.base_models_metadata_exists = True

    @property
    def base_models_metadata(self):
        if self.base_models_metadata_exists:
            return self.base_models_metadata_

        raise ValueError("Base Models given during initialization did not have any metadata stored!")

    def fit(self, X, y):
        """Fitting the ensemble. To do so, we get the predictions of the base models and pass it to the ensemble's fit.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,)
            The target values. An array of int.

        Returns:
        -------
        self : object
            Returns
        """
        if self.predict_method not in ["predict", "predict_proba"]:
            raise ValueError(f"Unknown predict method: {self.predict_method}")

        # Test if ensemble technique supports passthrough
        if self.passthrough and (not (hasattr(self, "supports_passthrough") and self.supports_passthrough is True)):
            raise ValueError("Passthrough is enabled but the ensemble technique does not support passthrough!")

        X, y = check_X_y(X, y)
        self.le_ = LabelEncoder().fit(y)
        self.classes_ = self.le_.classes_

        # Get the classes seen by the base model on the data they have been trained on.
        if hasattr(self.base_models[0], "le_"):
            self.base_model_le_ = self.base_models[0].le_

            # Check if self.classes_ differs
            if len(self.classes_) != len(self.base_model_le_.classes_):
                # TO fix it, we use the label encoder of the base models
                self.le_ = self.base_model_le_
                self.classes_ = self.base_model_le_.classes_

        y_ = self.le_.transform(y)

        if self.passthrough:
            self.ensemble_passthrough_fit(X, self.base_models_predictions(X), y_)
        else:
            self.ensemble_fit(self.base_models_predictions(X), y_)

        return self

    def predict(self, X):
        """Predicting with the ensemble. To do so, we get the predictions of the base models and pass it to the ensemble.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns:
        -------
        y : ndarray, shape (n_samples,)
            Vector containing the class labels for each sample.
        """
        check_is_fitted_for = []
        if hasattr(self, "fitted_attributes"):
            check_is_fitted_for = self.fitted_attributes

        check_is_fitted(self, ["le_", "classes_", *check_is_fitted_for])
        X = check_array(X)
        bm_preds = self.base_models_predictions_for_ensemble_predict(X)

        if self.passthrough:
            ensemble_output = self.ensemble_passthrough_predict(X, bm_preds)
        else:
            ensemble_output = self.ensemble_predict(bm_preds)

        return self.le_.inverse_transform(ensemble_output)

    def predict_proba(self, X):
        """Predicting with the ensemble.
         To do so, we get the predictions of the base models and pass it to the ensemble.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns:
        -------
        y : ndarray, shape (n_samples, n_classes)
            Vector containing the class probabilities for each class for each sample.
        """
        check_is_fitted(self, ["le_", "classes_"])
        X = check_array(X)
        bm_preds = self.base_models_predictions_for_ensemble_predict(X, predict_for="predict_proba")

        if self.passthrough:
            ensemble_output = self.ensemble_passthrough_predict_proba(X, bm_preds)
        else:
            ensemble_output = self.ensemble_predict_proba(bm_preds)

        return ensemble_output

    def base_models_predictions(self, X):
        if self.predict_method == "predict":
            return [self.le_.transform(bm.predict(X)) for bm in self.base_models]
        else:
            return [bm.predict_proba(X) for bm in self.base_models]

    @staticmethod
    def _confidences_to_predictions(confidences):
        return np.argmax(confidences, axis=1)

    def base_models_predictions_for_ensemble_predict(self, X, predict_for="predict"):
        # Catch dynamic case
        if self.predict_method_ensemble_predict == "dynamic":
            pred_method = predict_for
        else:
            pred_method = self.predict_method_ensemble_predict

        if pred_method == "predict":
            return [self.le_.transform(bm.predict(X)) for bm in self.base_models]
        else:
            return [bm.predict_proba(X) for bm in self.base_models]

    def oracle_predict(self, X, y):
        """Oracle Predict, predicting with knowing the ground truth. Only used by some methods for comparison.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.
        y : array-like, shape (n_samples,)
            The target values. An array of int.

        Returns:
        -------
        y : ndarray, shape (n_samples,)
            Vector containing the class labels for each sample.
        """
        X, y = check_X_y(X, y)
        self.classes_ = np.unique(y)
        y_ = self.le_.transform(y)

        ensemble_output = self.ensemble_oracle_predict(self.base_models_predictions_for_ensemble_predict(X), y_)

        return self.transform_ensemble_prediction(ensemble_output)

    @abstractmethod
    def ensemble_fit(self, base_models_predictions: list[np.ndarray], labels: np.ndarray) -> AbstractEnsemble:
        """Fit an ensemble given predictions of base models and targets.

        base_models_predictions can either be the raw predictions or the confidences!

        Parameters
        ----------
        base_models_predictions : array of shape = [n_base_models, n_data_points, n_targets]
            n_targets is the number of classes in case of classification,
            n_targets is 0 or 1 in case of regression

        labels : array of shape [n_targets]

        Returns:
        -------
        self

        """

    def ensemble_passthrough_fit(self, X, base_models_predictions: list[np.ndarray],
                                 labels: np.ndarray) -> AbstractEnsemble:
        """Fit an ensemble given predictions of base models and targets.

        base_models_predictions can either be the raw predictions or the confidences!

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.
        base_models_predictions : array of shape = [n_base_models, n_data_points, n_targets]
            n_targets is the number of classes in case of classification,
            n_targets is 0 or 1 in case of regression

        labels : array of shape [n_targets]

        Returns:
        -------
        self

        """
        raise NotImplementedError("ensemble_passthrough_fit is not implemented by every ensemble method!")

    @abstractmethod
    def ensemble_predict(self, base_models_predictions: list[np.ndarray]) -> np.ndarray:
        """Create ensemble predictions from the base model predictions.

        base_models_predictions can either be the raw predictions or the confidences!

        Parameters
        ----------
        base_models_predictions : array of shape = [n_base_models, n_data_points, n_targets]
            n_targets is the number of classes in case of classification,
            n_targets is 0 or 1 in case of regression

        Returns:
        -------
        y : ndarray, shape (n_samples,)
            Vector containing the class labels for each sample.
        """

    def ensemble_predict_proba(self, base_models_predictions: list[np.ndarray]) -> np.ndarray:
        """Create ensemble probability predictions from the base model predictions.

        base_models_predictions can either be the raw predictions or the confidences!

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.
        base_models_predictions : array of shape = [n_base_models, n_data_points, n_targets]
            n_targets is the number of classes in case of classification

        Returns:
        -------
        y : ndarray, shape (n_samples, n_targets)
            Vector containing the class probabilities for each sample.
        """
        raise NotImplementedError("ensemble_predict_proba is not implemented by every ensemble method!")

    def ensemble_passthrough_predict(self, X, base_models_predictions: list[np.ndarray]) -> np.ndarray:
        """Create ensemble predictions from the base model predictions.

        base_models_predictions can either be the raw predictions or the confidences!

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.
        base_models_predictions : array of shape = [n_base_models, n_data_points, n_targets]
            n_targets is the number of classes in case of classification,
            n_targets is 0 or 1 in case of regression

        Returns:
        -------
        y : ndarray, shape (n_samples,)
            Vector containing the class labels for each sample.
        """
        raise NotImplementedError("ensemble_passthrough_predict is not implemented by every ensemble method!")

    def ensemble_passthrough_predict_proba(self, X, base_models_predictions: list[np.ndarray]) -> np.ndarray:
        """Create ensemble probability predictions from the base model predictions.

        base_models_predictions can either be the raw predictions or the confidences!

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.
        base_models_predictions : array of shape = [n_base_models, n_data_points, n_targets]
            n_targets is the number of classes in case of classification

        Returns:
        -------
        y : ndarray, shape (n_samples, n_targets)
            Vector containing the class probabilities for each sample.
        """
        raise NotImplementedError("ensemble_passthrough_predict_proba is not implemented by every ensemble method!")

    def ensemble_oracle_predict(self, base_models_predictions: list[np.ndarray], labels: np.ndarray) -> np.ndarray:
        """Predict with an oracle-like ensemble given predictions of base models and targets.

        base_models_predictions can either be the raw predictions or the confidences!

        Parameters
        ----------
        base_models_predictions : array of shape = [n_base_models, n_data_points, n_targets]
            n_targets is the number of classes in case of classification,
            n_targets is 0 or 1 in case of regression

        labels : array of shape [n_targets]

        Returns:
        -------
        y : ndarray, shape (n_samples,)
            Vector containing the class labels for each sample.
        """
        raise NotImplementedError("ensemble_oracle_predict is not implemented by every ensemble method!")



"""Selection of Pre-defined behavior spaces."""

from __future__ import annotations


def get_bs_configspace_similarity_and_loss_correlation():
    # "bs_configspace_similarity_and_loss_correlation"
    from phem.methods.ensemble_selection.qdo.behavior_functions.basic import LossCorrelationMeasure
    from phem.methods.ensemble_selection.qdo.behavior_functions.implicit_diversity_metrics import (
        ConfigSpaceGowerSimilarity,
    )
    from phem.methods.ensemble_selection.qdo.behavior_space import BehaviorSpace

    return BehaviorSpace([ConfigSpaceGowerSimilarity, LossCorrelationMeasure])


def get_bs_ensemble_size_and_loss_correlation():
    # "bs_configspace_similarity_and_loss_correlation"
    from phem.methods.ensemble_selection.qdo.behavior_functions.basic import (
        EnsembleSize,
        LossCorrelationMeasure,
    )
    from phem.methods.ensemble_selection.qdo.behavior_space import BehaviorSpace

    return BehaviorSpace([EnsembleSize, LossCorrelationMeasure])



from __future__ import annotations

import copy
import sys
from collections.abc import Callable
from enum import Enum

import numpy as np
from ribs.archives._add_status import AddStatus
from ribs.emitters._emitter_base import EmitterBase
from scipy.special import softmax


class DiscreteWeightSpaceEmitter(EmitterBase):
    """An emitter for ensemble selection that searches in the discrete weight space.

    The emitter mutates or uses crossover such that it changes the discrete weight vector (similar to greedy
    ensemble selection).
    As a result of only changing the discrete weight vector in an appropriate way, the weights produce by this emitter
    always sum to 1 and for all weights w_i it holds 0 < w_i <= 1.

    Parameters
    ----------
    archive: ArchiveBase
        pyribs like archive
    n_base_models: int
        Number of base models.
    batch_size: int
        Number of solutions that the emitter shall propose.
    start_weight_vectors: List[Tuple[np.array (n_base_models,), int]]
        A list of start weight vectors and sample sizes to consider for initial mutations/crossover.
    starting_step_size: int, default=1
        Starting step size used to adjust weight vectors.
    elite_selection_method: str in {"deterministic", "combined_dynamic", "tournament"}, default="combined_dynamic"
        Method used to select the next elite(s) for mutation/crossover.
        *   deterministic: Select the best performing elite as next elite
        *   combined_dynamic: Combination of stochastic and deterministic. Initial 50% chance of either, later on
            the % is dynamically updated based on the performance of the exploration vs exploitation.
        *  tournament: Select the best performing N elites from a tournament of size 10.
           For tournament selection, we first randomly sample 10 elites from the archive. Then, we select the elite
           that wins the tournament. The tournament is a simple tournament where each elite is compared to another.
           For crossover, N=2. Otherwise, N=1. If less than 10 elites are in the archive, we use all existing
           and random initial solutions for the tournament.
    crossover: Optional[str] in {"average", "two_point_crossover"}, default="average"
        If not none, str described the method used for crossover. This works as follows: first we select two parents
        using elite_selection_method. Next, do crossover. Finally, we mutate the new weight vector by taking a step in
        the discrete weights space (We mutate depending on mutation_probability_after_crossover).
        The crossover strategies are all in the the discrete weight space and respect the weight constraints.
        *   average: The new weight vector is the average of the two parents.
        *   two_point_crossover: Two points are selected and children are produced by crossing over at selected points.
    crossover_probability: Union[float, Callable], default=0.5
        If crossover is not None, this determines how likely crossover of children is.
        Allowed values are:
            * Floats between 0 < crossover_probability <= 1.
            * Functions that return the crossover portability (float) for the current internal iteration of the emitter.
        Internally, all values are converted to a function that returns the mutation probability.
    crossover_probability_dynamic: bool, default=True
        If True, we update the crossover probability over time dynamically based on the performance of crossover /
        no crossover. For this to work, `crossover_probability` must be a float. To have the performance of both
        choices, 0 < crossover_probability < 1.
    mutation_probability_after_crossover: Union[float, Callable], default=0.5
        If crossover is not None, this determines how likely mutation of children found with crossover is.
        Allowed values are:
            * Floats between 0 <= mutation_probability_after_crossover <= 1.
                If the value is 0, we employ an emergency mutation rate. That is, if crossover did not
                produce any new offspring too many times (n=50), then we increase the mutation rate by 0.1 such that new
                children might be found. This is reset everytime ask() is called.
                If the value is not 0, too many rejections are handled by the emergency step size.
            * Functions that return the mutation portability (float) for the current internal iteration of the emitter.
        Internally, all values are converted to a function that returns the mutation probability.
    mutation_probability_after_crossover_dynamic: bool, default=True
        If True, we update the probability over time dynamically based on the performance of mutation /
        no mutation. For this to work, `mutation_probability_after_crossover` must be a float. To have the performance
        of both choices, 0 < mutation_probability_after_crossover < 1.
    shared_seen_resources: Tuple(set,dict), default=None
        If not None, we assume this to be shared objects across multiple simple step emitters to avoid proposing
        the same solutions.  resources
    dynamic_updates_consider_rejections: bool, default=False
        If true, the dynamic updates for combined_dynamic, mutation_probability_after_crossover_dynamic, or
        crossover_probability_dynamic take the number of rejections into account.
    bounds: default=None
        Bounds of the solution space. Emitter should respect these. None means no bounds.
        TODO support this parameter. By default the DiscreteWeightSpaceEmitter is bounded in [0,1]
    seed:
        Random seed use by emitter.
        TODO: add documentation for seed, currently assumes that integer seed is given
    """

    allowed_elite_selection_methods = {"deterministic", "tournament", "combined_dynamic"}

    def __init__(
        self,
        archive,
        n_base_models: int,
        start_weight_vectors: list[tuple[np.array, int]],
        batch_size: int,
        starting_step_size: int = 1,
        elite_selection_method: str = "combined_dynamic",
        crossover: str | None = "average",
        crossover_probability: float | Callable = 0.5,
        crossover_probability_dynamic: bool = True,
        mutation_probability_after_crossover: float | Callable = 0.5,
        mutation_probability_after_crossover_dynamic: bool = True,
        negative_steps: bool = False,
        shared_seen_resources: tuple[set, dict] | None = None,
        weight_random_elite_selection: bool = False,
        weight_random_step_selection: bool = False,
        dynamic_updates_consider_rejections: bool = False,
        bounds=None,
        seed=None,
    ):
        # Input Validation (can not do at "fit" time like in sklearn)
        if starting_step_size < 1:
            raise ValueError("starting_step_size must be larger than 1!")
        if elite_selection_method not in self.allowed_elite_selection_methods:
            raise ValueError(
                f"Unknown elite_selection_method! Allowed are: {self.allowed_elite_selection_methods}. Got: {elite_selection_method}."
            )

        self._rng = np.random.default_rng(seed)
        self._batch_size = batch_size
        self.n_base_models = n_base_models
        self.starting_step_size = starting_step_size
        self._emergency_step_size = 0
        self.crossover_reject_counter = 0
        self._emergency_mutation_rate = 0
        self._elite_selection_method = elite_selection_method
        self.negative_steps = negative_steps
        self.crossover = crossover

        # -- Handle mutation and crossover probabilities input
        if (not isinstance(mutation_probability_after_crossover, float)) and (
            not isinstance(mutation_probability_after_crossover, Callable)
        ):
            raise ValueError(
                "mutation_probability_after_crossover must be float or  Callable",
                f"Got: {type(mutation_probability_after_crossover)}",
            )
        if isinstance(mutation_probability_after_crossover, float) and (
            (mutation_probability_after_crossover < 0) or (mutation_probability_after_crossover > 1)
        ):
            raise ValueError(
                "for mutation_probability_after_crossover it must hold that hold that 0 <= mutation_probability_after_crossover <=1",
                f"Got: {mutation_probability_after_crossover}",
            )
        if isinstance(mutation_probability_after_crossover, float):
            self._mutation_probability_after_crossover = (
                lambda x: mutation_probability_after_crossover
            )
        else:
            self._mutation_probability_after_crossover = mutation_probability_after_crossover

        if (not isinstance(crossover_probability, float)) and (
            not isinstance(crossover_probability, Callable)
        ):
            raise ValueError(
                "crossover_probability must be float or  Callable",
                f"Got: {type(crossover_probability)}",
            )
        if isinstance(crossover_probability, float) and (
            (crossover_probability <= 0) or (crossover_probability > 1)
        ):
            raise ValueError(
                "for crossover_probability it must hold that hold that 0 < crossover_probability <=1",
                f"Got: {crossover_probability}",
            )
        if isinstance(crossover_probability, float):
            self._crossover_probability = lambda x: crossover_probability
        else:
            self._crossover_probability = crossover_probability

        if crossover_probability_dynamic and (not isinstance(crossover_probability, float)):
            raise ValueError(
                "crossover_probability_dynamic can only be used if crossover_probability is a float!"
            )
        self._crossover_probability_dynamic = crossover_probability_dynamic

        if mutation_probability_after_crossover_dynamic and (
            not isinstance(mutation_probability_after_crossover, float)
        ):
            raise ValueError(
                "_mutation_probability_after_crossover_dynamic can only be "
                + "used if mutation_probability_after_crossover is a float!"
            )
        self._mutation_probability_after_crossover_dynamic = (
            mutation_probability_after_crossover_dynamic
        )

        # Init super class
        EmitterBase.__init__(
            self,
            archive,
            n_base_models,
            bounds,
        )

        # -- Get Start weight vectors (all but one base model have weight zero for all base models)
        self._start_weight_vectors = start_weight_vectors
        self._original_start_weight_vectors = self._start_weight_vectors[:]  # copy

        # -- Set shared resources if needed
        if shared_seen_resources is not None:
            self._seen_percentages = shared_seen_resources[0]
            self._remaining_steps_for_hash = shared_seen_resources[1]
            for w, _ in self._start_weight_vectors:
                self._seen_percentages.add(hash(tuple(w)))
        else:
            self._seen_percentages = {hash(tuple(w)) for w, _ in self._start_weight_vectors}
            self._remaining_steps_for_hash = {}

        # -- Default vars for explore/exploit trade-off (only used if elite_selection_method=dynamic)
        self.explore = 0
        self.exploit = 0
        self.n_mutate = 0
        self.n_crossover = 0
        self._total_rejects = 0
        self._total_crossover_rejects = 0
        self._tmp_crossover_mutation_reject = 0
        self._crossover_mutation_reject = 0
        self._tmp_no_crossover_mutation_reject = 0
        self._no_crossover_mutation_reject = 0
        self._elite_origin_reject_counter = {k: 0 for k in EliteOrigins}
        self._tmp_elite_origin_reject_counter = None

        # -- Weighted Randomness (not used)
        self._weight_random_elite_selection = weight_random_elite_selection
        self._weight_random_step_selection = (
            weight_random_step_selection  # weight_random_step_selection
        )
        if self._weight_random_step_selection:
            np.seterr(invalid="ignore")  # since we abuse dividing by zero to obtain nan values.
            self._base_model_sample_weights = np.full(
                n_base_models, 1 / n_base_models
            )  # step direction weights

        # -- Analysis Stats
        self._n_ensemble_sizes = [0] * (n_base_models + 1)
        self.add_status_counts = [0, 0, 0]  # Not Added, improved, new

        # Counts how often ask() was called (while tell() is also part of an iteration, we ignore it here)
        self._internal_iterations_counter = 0
        self._mutation_probabilities_after_crossover_over_time = []
        self._crossover_probabilities_over_time = []
        self._random_elite_selection_probabilities_over_time = []

        # -- Stuff for Dynamic Updates
        self._origin_performance_template = {
            SolutionOrigins.no_crossover: [
                0,
                0,
                0,
            ],  # Objective value, improvement count, total count
            SolutionOrigins.crossover: [0, 0, 0],
            EliteOrigins.deterministic: [0, 0, 0],
            EliteOrigins.stochastic: [0, 0, 0],
            SolutionOrigins.co_mutation: [0, 0, 0],
            SolutionOrigins.co_no_mutation: [0, 0, 0],
        }
        self._random_elite_selection_probability = 0.5
        self._origin_performance_over_time = []
        self._rejections_over_time = []
        self.dynamic_updates_consider_rejections = dynamic_updates_consider_rejections

    @property
    def mutation_probability_after_crossover(self):
        return self._mutation_probability_after_crossover(self._internal_iterations_counter)

    @property
    def crossover_probability(self):
        return self._crossover_probability(self._internal_iterations_counter)

    @property
    def proposes_negative_weights(self):
        return self.negative_steps

    # -- Ask / Tell code
    def ask(self):
        """Return potential solutions to evaluate.

        Proposes self._batch_size many solutions using the methods in following order:
            1. First it proposes the start weight vectors (created during initialization) as long as they still exist
            2. Next use the elite_selection_method to find a next elite to mutate.

        Added tricks:
            * Rejections: Previous proposed solutions are not proposed again if they are created as a result of a
                mutation.
            * Emergency Step Size: Adapt the step size if too many rejects happened during a single call of ask().
                The emergency step size is used during the mutation to reach regions of the weight space that might be
                further away from the current position. Thus, we might stop proposing already proposed vectors.

        """
        # Solutions to return
        w_vecs_to_return = []

        # Handle too many rejections
        max_number_rejects = 50
        reject_counter = 0

        # Analysis values
        self._tmp_crossover_rejects = 0
        self._tmp_crossover_mutation_reject = 0
        self._tmp_no_crossover_mutation_reject = 0
        self._tmp_elite_origin_reject_counter = {k: 0 for k in EliteOrigins}

        # Emergency values
        self.crossover_reject_counter = 0  # used for emergency mutation rate increase
        self._emergency_step_size = 0
        self._emergency_mutation_rate = (
            0  # Only use when cross is not None and mutation_probability_after_crossover is 0
        )
        use_crossover = self.crossover is not None

        # A step counter needed to discretize weight vectors correctly
        number_of_steps = []
        # An origin counter needed for analysis and dynamic updates
        solution_origins = []
        elite_origins = []

        # Fill solutions to return
        while len(w_vecs_to_return) < self._batch_size:
            # Reject fail save, change step size to escape problematic reject region
            if reject_counter >= max_number_rejects:
                reject_counter = 0
                self._emergency_step_size += 1

            if self.crossover_reject_counter >= max_number_rejects:
                # Unable to produce new children using crossover often enough, increase mutation rate
                self.crossover_reject_counter = 0
                self._emergency_mutation_rate += 1

            # Fill with start weight vectors first
            if self._start_weight_vectors:
                # -- No mutation or selection needed, we just want to evaluate all base models once.
                elite_float_w, sample_size = self._start_weight_vectors.pop()

                w_vecs_to_return.append(elite_float_w)

                # -- Add ask/tell state
                number_of_steps.append(sample_size)
                solution_origins.append(SolutionOrigins.initialization)
                elite_origins.append(EliteOrigins.initialization)
            else:
                mutated_children = self._get_mutated_children(use_crossover)

                # Loop over mutated children and add if needed
                for (
                    proposed_float_weights,
                    new_sample_size,
                    reject_flag,
                ), sol_org, e_org in mutated_children:
                    # Handle reject
                    if reject_flag:
                        # No need to continue as this is the end of the loop anyway
                        reject_counter += 1
                    # Handle insert into batch
                    elif len(w_vecs_to_return) < self._batch_size:
                        # Need to check if list is not too large
                        w_vecs_to_return.append(proposed_float_weights)

                        # -- Add ask/tell state
                        number_of_steps.append(new_sample_size)
                        solution_origins.append(sol_org)
                        elite_origins.append(e_org)

                    # Handle edge case
                    else:
                        # We do not add to the list anymore. Otherwise, we would create a too long list
                        # - Only a problem if crossover returns 2 solutions.

                        # If list got filled in the previous iteration of this loop, we need to correctly
                        # set what we have seen. Because now we might have seen stuff that never gets proposed.
                        pfw_hash = hash(tuple(proposed_float_weights))
                        self._seen_percentages.remove(pfw_hash)

        # -- Collect Meta Info
        self._total_rejects += reject_counter
        self._total_crossover_rejects += self._tmp_crossover_rejects
        self._crossover_mutation_reject += self._tmp_crossover_mutation_reject
        self._no_crossover_mutation_reject += self._tmp_no_crossover_mutation_reject
        self._elite_origin_reject_counter = {
            k: self._tmp_elite_origin_reject_counter[k] + self._elite_origin_reject_counter[k]
            for k in self._elite_origin_reject_counter
        }

        # -- Important return values to keep state inbetween ask/tell
        self.last_ask_sample_sizes = (
            number_of_steps  # Required to correctly transform continuous to discrete weights
        )
        self.last_solutions_origins = solution_origins
        self.last_elite_origins = elite_origins

        # -- Get some analysis stats
        for sol in w_vecs_to_return:
            self._n_ensemble_sizes[np.sum(sol > 0)] += 1

        self._mutation_probabilities_after_crossover_over_time.append(
            self.mutation_probability_after_crossover
        )
        self._crossover_probabilities_over_time.append(self.crossover_probability)
        self._random_elite_selection_probabilities_over_time.append(
            self._random_elite_selection_probability
        )
        self._internal_iterations_counter += 1

        return np.array(w_vecs_to_return)

    def tell(self, solutions, objective_values, behavior_values, metadata=None):
        """Gather results from tell.

        Here, an important different is that we add the sample size to the archive as metadata such that we can
        correctly discretize weight vectors later. The self.last_ask_sample_sizes was set during the last call to ask().
        """
        if (metadata is not None) and (
            not np.array_equal(metadata, np.empty(len(solutions), dtype=object))
        ):
            raise ValueError("This emitter does not (yet) store external metadata.")

        loop_data = zip(
            solutions,
            objective_values,
            behavior_values,
            self.last_ask_sample_sizes,
            self.last_solutions_origins,
            self.last_elite_origins,
            strict=False,
        )

        tmp_origin_performance = copy.deepcopy(self._origin_performance_template.copy())

        for _i, (sol, obj, beh, meta, sol_origin, e_origin) in enumerate(loop_data):
            status, value = self.archive.add(sol, obj, beh, meta)

            # Store status for analysis
            self.add_status_counts[status] += 1

            # -- Collect data about Origin Performance
            if sol_origin != SolutionOrigins.initialization:
                tmp_origin_performance[sol_origin][0] += obj
                if status == AddStatus.IMPROVE_EXISTING:
                    tmp_origin_performance[sol_origin][1] += 1
                tmp_origin_performance[sol_origin][2] += 1

            if self._elite_selection_method == "combined_dynamic":
                if e_origin in [EliteOrigins.deterministic, EliteOrigins.stochastic]:
                    tmp_origin_performance[e_origin][0] += obj
                    if status == AddStatus.IMPROVE_EXISTING:
                        tmp_origin_performance[e_origin][1] += 1
                    tmp_origin_performance[e_origin][2] += 1

        # -- Update Origin summary where needed
        tmp_origin_performance[SolutionOrigins.crossover] = list(
            np.array(tmp_origin_performance[SolutionOrigins.co_mutation])
            + np.array(tmp_origin_performance[SolutionOrigins.co_no_mutation])
        )
        self._origin_performance_over_time.append(tmp_origin_performance)
        self._rejections_over_time.append(
            (
                self._tmp_crossover_rejects,
                self._tmp_crossover_mutation_reject,
                self._tmp_no_crossover_mutation_reject,
                self._tmp_elite_origin_reject_counter,
            )
        )

        # -- Only keep origin performance of last N iterations
        self._origin_performance = copy.deepcopy(self._origin_performance_template)
        self._rejections_counts = (0, 0, 0, {k: 0 for k in EliteOrigins})
        N = -10  # last 10 iterations
        for o_p in self._origin_performance_over_time[N:]:
            for k in self._origin_performance:
                self._origin_performance[k] = list(
                    np.array(self._origin_performance[k]) + np.array(o_p[k])
                )

        # Accumulate rejections
        for rejcs in self._rejections_over_time[N:]:
            tmp_val = list(np.array(self._rejections_counts[:-1]) + np.array(rejcs[:-1]))
            tmp_val_2 = {
                k: rejcs[-1][k] + self._rejections_counts[-1][k]
                for k in self._rejections_counts[-1]
            }
            tmp_val.append(tmp_val_2)
            self._rejections_counts = tmp_val

        # -- Update functions
        self._update_crossover_probability()
        self._update_mutation_probability_after_crossover()
        self._update_random_elite_selection_probability()
        self._update_base_model_sample_weights(solutions, objective_values)

    # -- Dynamic Update Functions
    def _update_crossover_probability(self):
        # "> 2" to guarantee that we are after the initial solution
        if (
            self._crossover_probability_dynamic
            and (self._origin_performance[SolutionOrigins.crossover][2] > 2)
            and (self._origin_performance[SolutionOrigins.no_crossover][2] > 2)
        ):
            # Consider reject rate in performance assessment
            if self.dynamic_updates_consider_rejections:
                reject_ratio_crossover = self._origin_performance[SolutionOrigins.crossover][-1] / (
                    self._origin_performance[SolutionOrigins.crossover][-1]
                    + self._rejections_counts[0]
                    + self._rejections_counts[1]
                )
                reject_ratio_no_crossover = self._origin_performance[SolutionOrigins.no_crossover][
                    -1
                ] / (
                    self._origin_performance[SolutionOrigins.no_crossover][-1]
                    + self._rejections_counts[2]
                )

                if self._origin_performance[SolutionOrigins.crossover][0] < 0:
                    reject_ratio_crossover = 1 + (1 - reject_ratio_crossover)

                if self._origin_performance[SolutionOrigins.no_crossover][0] < 0:
                    reject_ratio_no_crossover = 1 + (1 - reject_ratio_no_crossover)

                self._origin_performance[SolutionOrigins.crossover][0] *= reject_ratio_crossover
                self._origin_performance[SolutionOrigins.no_crossover][0] *= (
                    reject_ratio_no_crossover
                )

            # Update
            cr_prb = _avg_obj_update(
                self._origin_performance[SolutionOrigins.crossover],
                self._origin_performance[SolutionOrigins.no_crossover],
            )

            self._crossover_probability = lambda x: max(min(cr_prb, 0.9), 0.1)

    def _update_mutation_probability_after_crossover(self):
        if (
            self._mutation_probability_after_crossover_dynamic
            and (self._origin_performance[SolutionOrigins.co_mutation][2] > 2)
            and (self._origin_performance[SolutionOrigins.co_no_mutation][2] > 2)
        ):
            # Consider reject rate in performance assessment
            if self.dynamic_updates_consider_rejections:
                reject_ratio_co_mutation = self._origin_performance[SolutionOrigins.co_mutation][
                    -1
                ] / (
                    self._origin_performance[SolutionOrigins.co_mutation][-1]
                    + self._rejections_counts[1]
                )
                reject_ratio_co_no_mutation = self._origin_performance[
                    SolutionOrigins.co_no_mutation
                ][-1] / (
                    self._origin_performance[SolutionOrigins.co_no_mutation][-1]
                    + self._rejections_counts[0]
                )

                if self._origin_performance[SolutionOrigins.co_mutation][0] < 0:
                    reject_ratio_co_mutation = 1 + (1 - reject_ratio_co_mutation)

                if self._origin_performance[SolutionOrigins.co_no_mutation][0] < 0:
                    reject_ratio_co_no_mutation = 1 + (1 - reject_ratio_co_no_mutation)

                self._origin_performance[SolutionOrigins.co_mutation][0] *= reject_ratio_co_mutation
                self._origin_performance[SolutionOrigins.co_no_mutation][0] *= (
                    reject_ratio_co_no_mutation
                )

            cr_prb = _avg_obj_update(
                self._origin_performance[SolutionOrigins.co_mutation],
                self._origin_performance[SolutionOrigins.co_no_mutation],
            )

            self._mutation_probability_after_crossover = lambda x: max(min(cr_prb, 0.9), 0.1)

    def _update_random_elite_selection_probability(self):
        if (
            (self._elite_selection_method == "combined_dynamic")
            and (self._origin_performance[EliteOrigins.stochastic][2] > 2)
            and (self._origin_performance[EliteOrigins.deterministic][2] > 2)
        ):
            # Consider reject rate in performance assessment
            if self.dynamic_updates_consider_rejections:
                reject_ratio_stochastic = self._origin_performance[EliteOrigins.stochastic][-1] / (
                    self._origin_performance[EliteOrigins.stochastic][-1]
                    + self._rejections_counts[-1][EliteOrigins.stochastic]
                )

                reject_ratio_deterministic = self._origin_performance[EliteOrigins.deterministic][
                    -1
                ] / (
                    self._origin_performance[EliteOrigins.deterministic][-1]
                    + self._rejections_counts[-1][EliteOrigins.deterministic]
                )

                if self._origin_performance[EliteOrigins.stochastic][0] < 0:
                    reject_ratio_stochastic = 1 + (1 - reject_ratio_stochastic)

                if self._origin_performance[EliteOrigins.deterministic][0] < 0:
                    reject_ratio_deterministic = 1 + (1 - reject_ratio_deterministic)

                self._origin_performance[EliteOrigins.stochastic][0] *= reject_ratio_stochastic
                self._origin_performance[EliteOrigins.deterministic][0] *= (
                    reject_ratio_deterministic
                )

            rng_prb = _avg_obj_update(
                self._origin_performance[EliteOrigins.stochastic],
                self._origin_performance[EliteOrigins.deterministic],
            )

            self._random_elite_selection_probability = max(min(rng_prb, 0.95), 0.05)

    def _update_base_model_sample_weights(self, solutions, objective_values):
        if self._weight_random_step_selection:
            raise NotImplementedError

            baseline_performance = (
                np.max(objective_values) if self.archive.empty else self.archive.stats.obj_max
            )

            # Compute the contribution to the performance for each base model; scale it by how much it contributed
            avg_perf_per_bm = np.divide(
                (solutions * objective_values.reshape(-1, 1)).sum(axis=0), np.sum(solutions, axis=0)
            )
            # Read as: The base model's contribution was X times worse than the baseline on average
            rel_dist_to_baseline = abs(avg_perf_per_bm - baseline_performance) / abs(
                baseline_performance
            )

            # FIXME: fix this to be something smarter
            # 1 is a constant factor that is subtracted from the weights; smaller distances to baseline makes
            #   the subtracted value smaller
            update_per_weight = 1 * np.nan_to_num(rel_dist_to_baseline, nan=1)
            # Update and re-normalize
            self._base_model_sample_weights = softmax(
                self._base_model_sample_weights - update_per_weight
            )

    # -- Wrapper Function for Selection, Crossover, and mutation
    def _get_mutated_children(self, use_crossover):
        # -- Set where to sample from
        # Since rng random produces [0, 1), "<" is correct such that for 1 it is True and for 0% it is false
        do_crossover = use_crossover and (self._rng.random() < self.crossover_probability)

        # -- Check if crossover is used
        if do_crossover:
            parents = self.get_next_elites(n_elites=2)

            # Handle origins (if two different origins, use combined origin, which can only be stoch+deter)
            parent_origins = [o for _, o in parents]
            if (EliteOrigins.stochastic in parent_origins) and (
                EliteOrigins.deterministic in parent_origins
            ):
                elite_origin = EliteOrigins.stochastic_and_deterministic
            else:
                elite_origin = parent_origins[0]
            sol_origin = None

            parent_elites = [e for e, _ in parents]
            children = self._crossover_elites(parent_elites[0], parent_elites[1])
            tmp_mutated_children = []
            tmp_elite_origins = []

            for c_elite_float_w, c_sample_size in children:
                # -- Decide if mutation is used
                do_mutation_after_crossover = self._rng.random() < (
                    self.mutation_probability_after_crossover + self._emergency_mutation_rate
                )

                # -- Do mutation or not
                if do_mutation_after_crossover:
                    tmp_mutated_children.append(self._mutate_elite(c_elite_float_w, c_sample_size))
                    tmp_elite_origins.append(elite_origin)
                    sol_origin = SolutionOrigins.co_mutation

                    # -- Track numbers
                    int_reject = int(tmp_mutated_children[-1][-1])
                    self._tmp_crossover_mutation_reject += int_reject
                    self._tmp_elite_origin_reject_counter[elite_origin] += int_reject

                else:
                    # -- Rejection Handling for crossover
                    pfw_hash = hash(tuple(c_elite_float_w))
                    if pfw_hash in self._seen_percentages:
                        self.crossover_reject_counter += 1
                        self._tmp_crossover_rejects += 1
                        self._tmp_elite_origin_reject_counter[elite_origin] += 1
                    else:
                        self._seen_percentages.add(pfw_hash)
                        tmp_mutated_children.append((c_elite_float_w, c_sample_size, False))
                        tmp_elite_origins.append(elite_origin)
                        sol_origin = SolutionOrigins.co_no_mutation

        # -- Default to mutation otherwise
        else:
            (elite_float_w, sample_size), elite_origin = self.get_next_elites()[0]
            tmp_mutated_children = [self._mutate_elite(elite_float_w, sample_size)]
            tmp_elite_origins = [elite_origin]
            sol_origin = SolutionOrigins.no_crossover

            # -- Track numbers
            int_reject = int(tmp_mutated_children[-1][-1])
            self._no_crossover_mutation_reject += int_reject
            self._tmp_elite_origin_reject_counter[elite_origin] += int_reject

        # -- Add flag (extra iteration most simple implementation)
        mutated_children = []
        for m_c, elite_origin in zip(tmp_mutated_children, tmp_elite_origins, strict=False):
            mutated_children.append((m_c, sol_origin, elite_origin))

        return mutated_children

    # -- Selection + Exploration/Exploitation Code
    def _get_random_elite(self):
        if self._weight_random_elite_selection:
            # Performance-weighted Random
            #   First implementation.
            elites = list(self.archive)
            if len(elites) == 1:
                elite_float_w, _, _, _, sample_size = elites[0]
            else:
                # Get proba_weights based on performances
                performances = np.array([e.obj for e in self.archive])
                sel_idx = self._rng.choice(
                    len(elites), p=np.exp(performances) / np.exp(performances).sum()
                )
                elite_float_w, _, _, _, sample_size = elites[sel_idx]
        else:
            # Uniform Random
            elite_float_w, _, _, _, sample_size = self.archive.get_random_elite()
        return elite_float_w, sample_size

    def _get_deterministic_elite(self, n=1):
        # Take the best elite
        performances = [e.obj for e in self.archive]
        elites = list(self.archive)

        if n == 1:
            i_best = np.argmax(performances)
            elite_float_w = elites[i_best].sol
            sample_size = elites[i_best].meta
            return [(elite_float_w, sample_size)]
        else:
            return [
                (elites[i_best].sol, elites[i_best].meta)
                for i_best in np.argpartition(performances, -n)[-n:]
            ]

    def _edge_case_select_next_elite(self, n_elites=1):
        # Randomly select something from the original start weight vector for mutation
        #   Choice workaround due to deprecated ndarray creation inside of choice.

        # -- Edge case where we need to create new weight vectors
        if n_elites > len(self._original_start_weight_vectors):
            # Get all existing
            elites = self._original_start_weight_vectors[:]  # copy

            # while due to possible rejection
            while len(elites) < n_elites:
                # Get list of possible start weight vector to mutate
                sel_idx = self._rng.choice(
                    len(self._original_start_weight_vectors),
                    min(n_elites - len(elites), len(self._original_start_weight_vectors)),
                    replace=False,
                )
                for ix in sel_idx:
                    # Mutate and add to list
                    *sampled_elite, reject_flag = self._mutate_elite(
                        *self._original_start_weight_vectors[ix]
                    )
                    if not reject_flag:
                        elites.append(sampled_elite)

            # Shuffle to remove order
            elites = [elites[ix] for ix in self._rng.permutation(len(elites))]

        # -- Default case
        else:
            sel_idx = self._rng.choice(
                len(self._original_start_weight_vectors), n_elites, replace=False
            )
            elites = [self._original_start_weight_vectors[ix] for ix in sel_idx]

        return elites

    def _select_next_elite(self, return_origin=False):
        if self._elite_selection_method in ["combined_dynamic"]:
            if self._rng.random() < self._random_elite_selection_probability:
                self.explore += 1
                origin = EliteOrigins.stochastic
                elite_float_w, sample_size = self._get_random_elite()
            else:
                self.exploit += 1
                origin = EliteOrigins.deterministic
                elite_float_w, sample_size = self._get_deterministic_elite()[0]
        else:
            raise ValueError("Unknown elite selection method! Got:", self._elite_selection_method)

        if return_origin:
            return (elite_float_w, sample_size), origin

        return elite_float_w, sample_size

    def _edge_case_selection(self, n_elites) -> list[tuple[np.ndarray, int]]:
        elites = []

        # Handle edge case where this is the first call of ask() and the archive is still empty, but
        # we already proposed all start weight vectors.
        # This happens if batch size > len(self._start_weight_vectors).
        if self.archive.stats.num_elites == 0:
            elites = self._edge_case_select_next_elite(n_elites)
        elif (n_elites > 1) and (self.archive.stats.num_elites < n_elites):
            elites = self._get_deterministic_elite(self.archive.stats.num_elites)
            rng_elites = self._edge_case_select_next_elite(n_elites - self.archive.stats.num_elites)
            elites.extend(rng_elites)

        return elites

    def _tournament_selection(self, n_elites):
        # Following https://en.wikipedia.org/wiki/Tournament_selection
        n_competitors = n_elites * 10
        p = 0.8
        p_elites = [p * ((1 - p) ** i) for i in range(n_competitors)]
        p_elites = np.array(p_elites) / np.sum(p_elites)  # Guarantee sum of probabilities is 1

        # -- Edge case not enough elites for real tournament
        # TODO: handle not enough elites differently
        elites = self._edge_case_selection(n_competitors)

        if not elites:
            # -- Tournament
            all_elites = list(self.archive)
            # Select without replacement
            elites = [
                all_elites[j]
                for j in self._rng.choice(len(all_elites), n_competitors, replace=False)
            ]
            perf_elites = [e.obj for e in elites]
            elites = np.array(
                [(elite_float_w, sample_size) for elite_float_w, _, _, _, sample_size in elites],
                dtype="object",
            )
            elites = elites[np.argsort(-np.array(perf_elites))]  # Sort by performance
            elites = [tuple(x) for x in elites]

        # Transform into specific format
        sel_idx = self._rng.choice(n_competitors, n_elites, p=p_elites, replace=False)
        return [tuple(elites[idx]) for idx in sel_idx]

    def get_next_elites(self, n_elites=1):
        # -- Tournament is an extra case
        if self._elite_selection_method == "tournament":
            elites = self._tournament_selection(n_elites)
            return [(elite, EliteOrigins.tournament) for elite in elites]

        # -- Handle edge case
        elites = self._edge_case_selection(n_elites)
        if elites:
            return [(elite, EliteOrigins.edge_case) for elite in elites]

        # -- Non Edge Case
        # Special case
        if self._elite_selection_method == "deterministic":
            # Edge case to avoid same deterministic elite everytime
            self.exploit += n_elites
            elites = self._get_deterministic_elite(n=n_elites)

            return [(elite, EliteOrigins.deterministic) for elite in elites]

        # Default case
        return [self._select_next_elite(return_origin=True) for _ in range(n_elites)]

    # -- Crossover
    def _crossover_elites(self, p1, p2):
        self.n_crossover += 1

        p1_elite_float_w, p1_sample_size = p1
        p2_elite_float_w, p2_sample_size = p2
        p1_elite_dis_w = _calculate_counts(p1_elite_float_w, p1_sample_size)
        p2_elite_dis_w = _calculate_counts(p2_elite_float_w, p2_sample_size)

        if self.crossover == "average":
            children = self._crossover_average(p1_elite_dis_w, p2_elite_dis_w)
        elif self.crossover == "two_point_crossover":
            children = self._crossover_two_point(p1_elite_dis_w, p2_elite_dis_w)
        else:
            raise ValueError(f"Unknown crossover strategy. Got: {self.crossover}")

        return children

    def _crossover_average(self, p1_elite_dis_w, p2_elite_dis_w):
        # Naive average crossover that returns ints
        proposed_discrete_weights = np.ceil((p1_elite_dis_w + p2_elite_dis_w) / 2)
        new_sample_size = int(sum(proposed_discrete_weights))
        proposed_float_weights = _calculate_weights(proposed_discrete_weights)
        return [(proposed_float_weights, new_sample_size)]

    def _crossover_two_point(self, p1_elite_dis_w, p2_elite_dis_w):
        non_zero_mask = (p1_elite_dis_w != 0) | (p2_elite_dis_w != 0)
        n_non_zero = sum(non_zero_mask)

        # Handle Edge Case
        if n_non_zero < 3:
            # Not enough unique ensemble member for two point, simply return average
            children = self._crossover_average(p1_elite_dis_w, p2_elite_dis_w)

        # Default Two Point
        else:
            # Find crossover points
            co_points = sorted(self._rng.choice(list(range(n_non_zero)), size=2, replace=False))
            org_p1 = np.copy(p1_elite_dis_w)
            org_p2 = np.copy(p2_elite_dis_w)
            child_1_to_fill = np.copy(p1_elite_dis_w[non_zero_mask])
            child_2_to_fill = np.copy(p2_elite_dis_w[non_zero_mask])

            # Get what to fill and fill it
            fill_p1 = p2_elite_dis_w[non_zero_mask][co_points[0] : co_points[1]]
            child_1_to_fill[co_points[0] : co_points[1]] = fill_p1

            fill_p2_1 = p1_elite_dis_w[non_zero_mask][: co_points[0]]
            fill_p2_2 = p1_elite_dis_w[non_zero_mask][co_points[1] :]
            child_2_to_fill[: co_points[0]] = fill_p2_1
            child_2_to_fill[co_points[1] :] = fill_p2_2

            # Create full children
            p1_elite_dis_w[non_zero_mask] = child_1_to_fill
            p2_elite_dis_w[non_zero_mask] = child_2_to_fill

            # Fill children list and make sure to catch bad mutations
            children = []
            if sum(p1_elite_dis_w) != 0:
                children.append((_calculate_weights(p1_elite_dis_w), int(sum(p1_elite_dis_w))))

            if sum(p2_elite_dis_w) != 0:
                children.append((_calculate_weights(p2_elite_dis_w), int(sum(p2_elite_dis_w))))

            # Edge case where nothing happened
            if not children:
                # Fall Back to average
                children = self._crossover_average(org_p1, org_p2)

        return children

    # -- Mutation
    def _mutate_elite(self, elite_float_w, sample_size):
        """Mutate existing elite to create a new elite.

        We mutate existing weight vectors by taking one or multiple steps in the discrete weight vector space.
        The discrete weight vector space consists of the number of times a base model was sampled.
        """
        self.n_mutate += 1

        # -- Current default method: mutate the discrete weight vector
        # Get discrete weight vector
        elite_dis_w = _calculate_counts(elite_float_w, sample_size)
        # Step (i.e., mutate) the weight vector
        proposed_discrete_weights, step_distance = self._step_for_weights(elite_dis_w)
        # Compute new weights
        proposed_float_weights = _calculate_weights(proposed_discrete_weights)

        # Verify new weight vector and reject if necessary, otherwise add to known weights
        pfw_hash = hash(tuple(proposed_float_weights))
        if pfw_hash in self._seen_percentages:
            return None, None, True
        self._seen_percentages.add(pfw_hash)

        # Employ correct step size here to represent samples correct and fix discretization error
        new_sample_size = sample_size + step_distance

        return proposed_float_weights, new_sample_size, False

    # - Step Selection Code
    def _update_step_lookup(self, idx_for_steps, step_size, dwv_hash):
        possible_steps = [(idx, step_size) for idx in idx_for_steps]

        if self.negative_steps:
            possible_steps.extend([(idx, -step_size) for idx in idx_for_steps])

        self._remaining_steps_for_hash[dwv_hash] = possible_steps

    def _init_step_lookup(self, discrete_weight_vector, dwv_hash):
        idx_for_steps = list(range(self.n_base_models))
        # Remove step for same base model if only 1 model seen so far
        if sum(discrete_weight_vector) == 1:
            idx_for_steps.remove(np.argmax(discrete_weight_vector))

        self._update_step_lookup(idx_for_steps, self.starting_step_size, dwv_hash)

    def _refill_step_lookup(self, discrete_weight_vector, dwv_hash, new_step_size):
        idx_for_steps = list(range(self.n_base_models))
        # Remove step for same base model if only one model has all steps so far
        if sum(discrete_weight_vector) == np.max(discrete_weight_vector):
            idx_for_steps.remove(np.argmax(discrete_weight_vector))

        self._update_step_lookup(idx_for_steps, new_step_size, dwv_hash)

    def _step_for_weights(self, discrete_weight_vector):
        """Main Code that does the steps for a given weight vector.

        Returns:
            new discrete weight vector
            step distance (absolute step size), also identical to how many "samples" were taken.
        """
        tmp_dmw = discrete_weight_vector.copy()

        # --- Setup hash and steps
        dwv_hash = hash(tuple(discrete_weight_vector))
        if dwv_hash not in self._remaining_steps_for_hash:
            self._init_step_lookup(discrete_weight_vector, dwv_hash)

        # --- Select a step direction (i.e., which base model's weight to change)
        if self._weight_random_step_selection:
            selected_step = tuple(
                self._rng.choice(
                    self._remaining_steps_for_hash[dwv_hash],
                    p=softmax(
                        [
                            self._base_model_sample_weights[idx]
                            for idx, _ in self._remaining_steps_for_hash[dwv_hash]
                        ]
                    ),
                )
            )
        else:
            selected_step = tuple(self._rng.choice(self._remaining_steps_for_hash[dwv_hash]))
        self._remaining_steps_for_hash[dwv_hash].remove(selected_step)
        step_direction, step_size = selected_step

        # --- Take step and handle emergency step size correctly
        if step_size < 0:
            tmp_step_size = step_size - self._emergency_step_size
        else:
            tmp_step_size = step_size + self._emergency_step_size

        tmp_dmw[step_direction] += tmp_step_size

        # --- Increase step size if finished
        if len(self._remaining_steps_for_hash[dwv_hash]) == 0:
            self._refill_step_lookup(discrete_weight_vector, dwv_hash, abs(tmp_step_size) + 2)

        return tmp_dmw, abs(tmp_step_size)


# -- Utility functions
def _calculate_weights(discrete_weights) -> np.ndarray:
    # abs() to handle negative discrete steps!
    return discrete_weights / np.abs(discrete_weights).sum()


def _calculate_counts(float_weights, sample_size) -> np.ndarray:
    return float_weights * sample_size


def _avg_obj_update(origin_performance_stats_1, origin_performance_stats_2):
    # Get avg obj score
    avg_rng_perf = origin_performance_stats_1[0] / origin_performance_stats_1[2]
    avg_deter_perf = origin_performance_stats_2[0] / origin_performance_stats_2[2]
    eps = sys.float_info.epsilon

    return 1 - (avg_rng_perf + eps) / (avg_rng_perf + avg_deter_perf + eps)


# -- Enums


class SolutionOrigins(Enum):
    initialization = (0,)
    no_crossover = (1,)
    crossover = (2,)
    co_mutation = (3,)
    co_no_mutation = (4,)


class EliteOrigins(Enum):
    initialization = (0,)
    tournament = (1,)
    edge_case = (2,)
    stochastic = (3,)
    deterministic = (4,)
    stochastic_and_deterministic = 5



from __future__ import annotations

import numpy as np
from ribs.archives._archive_base import ArchiveBase


class QualityArchive(ArchiveBase):
    """A baseline archive that is not a typical QDO archive.

    It functions like a pure quality population. Thus, allowing us to compare diversity inspired populations of elites
    to simple quality archives. It is a collection of N items w.r.t. some simple characteristics.
    Usually, it is the top N  performing solutions.

    We assume that we want to maximize the objective following the default QDO implementation.

    Parameters
    ----------
    archive_size: int
        Size of the beam.
    behavior_n_dim: int
        Number of dimensions of the behavior space. Needed to correctly store optional behavior values.
    seed: int
        Seed for the random number generator of archive base. (Beam is not random)
    """

    def __init__(self, archive_size, behavior_n_dim, seed, show_analysis=False):
        self._dims = [archive_size]
        self.behavior_n_dim = behavior_n_dim
        self.archive_size = archive_size
        self.stored_values_count = 0

        ArchiveBase.__init__(
            self,
            storage_dims=tuple(self._dims),
            behavior_dim=self.behavior_n_dim,
            seed=seed,
        )
        self._show_analysis = show_analysis
        self._elites_over_time = []

    def get_index(self, behavior_values):
        """Return the index of the position in the archive that should be replaced by the behavior values."""
        # Fill empty bin
        if len(self._occupied_indices) < self.archive_size:
            return tuple(np.unravel_index(np.argmin(self._occupied), self._occupied.shape))

        # Find bin with the smallest value and return (local competition will resolve the if it is inserted or not)
        return tuple(np.unravel_index(np.argmin(self._objective_values), self._objective_values.shape))

    def add(self, solution, objective_value, behavior_values, metadata=None):
        status, value = ArchiveBase.add(self, solution, objective_value, behavior_values, metadata)
        self.stored_values_count += 1

        if self._show_analysis and (self.stored_values_count % 50 == 0):
            self._elites_over_time.append([(e.obj, *e.beh) for e in self])

        return status, value



from __future__ import annotations

from functools import partial

from phem.base_utils.diversity_metrics import LossCorrelation
from phem.methods.ensemble_selection.qdo.behavior_space import BehaviorFunction


def ensemble_size(weights) -> float:
    return sum(weights != 0)

# -- Make Behavior Functions
# - Diversity Metrics
LossCorrelationMeasure = BehaviorFunction(partial(LossCorrelation, checks=False), ["y_true", "Y_pred_base_models"],
                                          (0, 1), "proba", name=LossCorrelation.name + "(Lower is more Diverse)")
EnsembleSize = BehaviorFunction(ensemble_size, ["weights"], (0, 50), "none", name="Ensemble Size")
